
import sys, os
from pathlib import Path

from omnibelt import cwd

import _test_util as tu
from _test_util import reset_profile


import omnifig as fig



def test_base_profile():
	profile = reset_profile()
	assert profile.name == 'base-profile'
	assert profile.path == tu.PROFILES_PATH / 'base.yaml'
	assert profile is fig.ProfileBase._profile



def test_missing_profile():
	profile = reset_profile('does-not-exist')
	assert profile.name == ''



def test_bad_profile():
	profile = reset_profile('bad')
	assert profile.name == ''



def test_current_project():
	profile = reset_profile()
	proj = fig.get_current_project()
	# assert proj.name == 'omni-fig' # inferred
	assert proj.name is None



def test_current_project2():
	profile = reset_profile()
	with cwd(Path('/')):
		proj = fig.get_current_project()
	assert proj.name is None



def test_load_project():
	profile = reset_profile()

	old = profile.get_current_project()

	proj = profile.switch_project('example1')

	new = profile.get_current_project()

	assert old is not None
	assert old is not new
	assert proj is new

	assert proj.name == 'project1' # name is specified in project info file

	assert not proj.is_activated
	proj.activate()
	assert proj.is_activated

	assert len(list(proj.iterate_configs())) == 1

	assert proj.create_config('just-one').pull('hello') == 'world'
	assert fig.create_config('just-one').pull('hello') == 'world'



def test_load_missing_project():
	profile = reset_profile()

	try:
		fig.get_project('does-not-exist')
	except profile.UnknownProjectError:
		pass
	else:
		assert False, 'should have raised an UnknownProjectError'



def test_new_profile():
	profile = reset_profile()
	fig.get_current_project()
	assert len(list(profile.iterate_projects())) == 1

	reset_profile()

	profile = fig.get_profile()
	assert len(list(profile.iterate_projects())) == 0



def test_ambiguous_project_name():
	profile = reset_profile()

	proj = profile.get_project('example1')
	assert not proj.is_activated

	assert proj.name == 'project1' # name is specified in project info file
	assert proj is profile.get_project('project1')
	assert proj is profile.get_project('example1')

	proj = profile.get_project('example2')
	assert not proj.is_activated

	assert proj.name == 'example2' # no name is specified in project info file
	assert proj is profile.get_project('example2')



def test_active_project():
	assert not getattr(fig, '_loaded_p2', False) # flag
	profile = reset_profile('active')
	assert getattr(fig, '_loaded_p2', False) # means p2 source code was run
	del fig._loaded_p2

	assert profile.name == 'active-profile'

	proj = fig.get_current_project()

	# assert proj.name == 'omni-fig'
	assert proj.name is None

	cmp = proj.find_component('cmp1', None)

	assert cmp is not None
	assert cmp.cls.__name__ == 'Cmpn1'
	assert cmp.cls.__module__ == 'my_code'
	assert cmp.project is not proj
	assert cmp.project is profile.get_project('example2')



def test_parent_is_project():
	profile = reset_profile('does-not-exist')

	with cwd(tu.PROJECTS_PATH / 'p8' / 'some-dir' / 'deep'):
		fig.initialize()

		assert fig.get_current_project().name == 'p8'



def test_explicit_project_file():

	profile = reset_profile('multi')

	proj = profile.get_current_project()
	# assert proj.name == 'omni-fig'
	assert proj.name is None

	proj3 = profile.get_project('example3')
	assert proj3.name == 'example3'

	assert len(list(profile.iterate_projects())) == 2

	proj3a = profile.get_project('example3-alt')
	assert proj3a.name == 'other-p3'

	assert len(list(profile.iterate_projects())) == 3

	assert not proj3.is_activated
	assert not getattr(fig, '_loaded_p3', False)

	proj3.activate()
	assert proj3.is_activated

	assert getattr(fig, '_loaded_p3_util', 0) == 1
	assert getattr(fig, '_loaded_p3_script', 0) == 1
	assert getattr(fig, '_loaded_p3_script2', 0) == 0

	assert len(list(proj3.iterate_components())) == 2

	assert sorted(entry.cls.__name__ for entry in proj3.iterate_components()) == ['P3C1', 'UtilComponent']

	proj3a.activate()

	assert getattr(fig, '_loaded_p3_util', 0) == 2
	assert getattr(fig, '_loaded_p3_script', 0) == 2
	assert getattr(fig, '_loaded_p3_script2', 0) == 1

	assert len(list(proj3a.iterate_components())) == 3

	assert sorted(entry.cls.__name__ for entry in proj3a.iterate_components()) == ['P3C1', 'P3aC1', 'UtilComponent']



def test_load_noncurrent_project(): # "hijack" project contents
	reset_profile()

	# default = fig.get_current_project()
	default = fig.get_profile().create_project('test-project')

	# assert default.name == 'omni-fig'
	assert default.name == 'test-project'
	assert not default.is_activated

	proj = fig.get_project('example2')

	# assert fig.get_current_project().name == 'omni-fig'
	assert fig.get_current_project().name == 'test-project'
	assert not default.is_activated

	proj.load()

	# assert fig.get_current_project().name == 'omni-fig'
	assert fig.get_current_project().name == 'test-project'
	assert not default.is_activated
	assert not proj.is_activated

	assert len(list(default.iterate_components())) == 4

	# assert fig.get_current_project().name == 'omni-fig'
	assert fig.get_current_project().name == 'test-project'
	assert default.is_activated
	assert not proj.is_activated

	proj.activate()

	assert len(list(proj.iterate_components())) == 4

	assert [e.cls.__name__ for e in proj.iterate_components()] \
	       == [e.cls.__name__ for e in default.iterate_components()]



def test_infer_project():

	profile = reset_profile()
	profile.data['projects'] = {name: str(Path(path).absolute()) for name, path in profile.data['projects'].items()}

	old = os.getcwd()
	os.chdir(str(tu.PROJECTS_PATH / 'p1' / 'some' / 'sub-dir'))

	proj = fig.get_current_project()

	assert proj.name == 'project1'

	os.chdir(old)



def test_xray():
	profile = reset_profile()

	proj = profile.get_project('example2')

	print()
	proj.xray('script')

	print()
	proj.xray('component')

	print()
	proj.xray('modifier')

	print()
	proj.xray('creator')

	print()
	proj.xray('config')

	assert len(proj.xray('script', as_list=True)) == 1
	assert len(proj.xray('component', as_list=True)) == 4
	assert len(proj.xray('modifier', as_list=True)) == 3
	assert len(proj.xray('creator', as_list=True)) == 0
	assert len(proj.xray('config', as_list=True)) == 3



def test_xray_related():
	profile = reset_profile('active')

	fig.initialize('example3')

	proj = profile.switch_project('example4')

	print()
	proj.xray('script')

	print()
	proj.xray('component')

	print()
	proj.xray('modifier')

	print()
	proj.xray('creator')

	print()
	proj.xray('config')

	assert len(proj.xray('script', as_list=True)) == 2
	assert len(proj.xray('component', as_list=True)) == 11
	assert len(proj.xray('modifier', as_list=True)) == 4
	assert len(proj.xray('creator', as_list=True)) == 1
	assert len(proj.xray('config', as_list=True)) == 3



def test_find_base_projects():
	reset_profile()

	fig.initialize('example2', 'example4')

	assert fig.get_current_project().name is None

	proj = fig.get_project()
	p2 = fig.get_project('example2')
	p4 = fig.get_project('example4')

	assert proj.find_component('cmp1').cls.__name__ == 'Cmpn1'
	assert proj.find_component('example4:cmp1').cls.__name__ == 'P4C1'
	assert p2.find_component('cmp1').cls.__name__ == 'Cmpn1'
	assert p4.find_component('cmp1').cls.__name__ == 'P4C1'

	@fig.component('cmp1')
	class MyCmpn1:
		pass

	assert proj.find_component('cmp1').cls.__name__ == 'MyCmpn1'
	assert proj.find_component('example2:cmp1').cls.__name__ == 'Cmpn1'
	assert proj.find_component('example4:cmp1').cls.__name__ == 'P4C1'
	assert p2.find_component('cmp1').cls.__name__ == 'Cmpn1'
	assert p4.find_component('cmp1').cls.__name__ == 'P4C1'

	assert proj.find_component('cmp100').cls.__name__ == 'P4C100'
	assert p2.find_component('cmp100').cls.__name__ == 'P4C100'

	assert proj.find_component('cmp2').cls.__name__ == 'Cmpn2'
	assert p2.find_component('cmp2').cls.__name__ == 'Cmpn2'
	assert p4.find_component('cmp2').cls.__name__ == 'Cmpn2'



def test_related():
	profile = reset_profile()
	assert len(list(profile.iterate_projects())) == 0

	fig.initialize('example2')
	assert len(list(profile.iterate_projects())) == 2

	assert fig.get_current_project().name is None

	proj = fig.switch_project('example4')
	assert len(list(profile.iterate_projects())) == 3

	proj.activate()

	assert len(list(profile.iterate_projects())) == 3
	assert fig.get_current_project() is proj
	assert proj.is_activated

	p1 = fig.get_project('example1')
	assert not p1.is_activated

	p2 = fig.get_project('example2')
	assert p2.is_activated

	assert len(list(profile.iterate_projects())) == 4
	assert 'example5' not in profile._loaded_projects

	try:
		proj.find_component('cmp6')
	except proj.UnknownArtifactError:
		pass
	else:
		assert False, 'expected exception'

	assert 'example5' in profile._loaded_projects
	assert len(list(profile.iterate_projects())) == 5

	p5 = fig.get_project('example5')

	assert p5.is_activated
	assert proj.find_component('cmp5').cls.__name__ == 'P5C5'

	assert proj.find_component('cmp3').cls.__name__ == 'P5C3'
	assert proj.find_component('example2:cmp3').cls.__name__ == 'Cmpn3'

	assert proj.find_component('cmp1').cls.__name__ == 'P4C1'
	assert proj.find_component('example2:cmp1').cls.__name__ == 'Cmpn1'

	assert proj.find_component('cmp4').cls.__name__ == 'Cmpn4'



def test_infer_project2():
	profile = reset_profile('does-not-exist')

	with cwd(tu.PROJECTS_PATH / 'p1'):
		profile = reset_profile()

		print()
		fig.main(['-h'])

		assert fig.get_current_project().name == 'project1'






from pathlib import Path

from _test_util import reset_profile, PROJECTS_PATH

import omnifig as fig



def test_initialize_empty():
	profile = reset_profile()

	default = fig.get_current_project()

	assert default.name is None
	assert not default.is_activated

	assert len(list(profile.iterate_projects())) == 1

	fig.initialize()

	assert len(list(profile.iterate_projects())) == 1
	assert default.is_activated

	assert default is fig.get_current_project()



def test_initialize():
	profile = reset_profile()

	default = fig.get_current_project()

	assert default.name is None
	assert not default.is_activated

	fig.initialize('example2')

	assert len(list(profile.iterate_projects())) == 2

	proj = fig.get_project('example2')

	assert proj.is_activated
	assert default.is_activated

	assert len(list(default.iterate_components())) == 0
	assert fig.get_current_project().name is None

	assert proj.name == 'example2'
	assert len(list(proj.iterate_components())) == 4



def test_initialize_multi():
	profile = reset_profile()

	default = fig.get_current_project()

	assert default.name is None
	assert not default.is_activated

	assert len(list(profile.iterate_projects())) == 1

	fig.initialize('example2', 'example4')

	assert len(list(profile.iterate_projects())) == 3

	assert default is fig.get_current_project()
	assert len(list(profile.iterate_projects())) == 3
	assert [True, True, True] == [p.is_activated for p in profile.iterate_projects()]

	assert len(list(default.iterate_components())) == 0

	proj = fig.get_project('example2')
	assert len(list(proj.iterate_components())) == 4

	proj = fig.get_project('example4')
	assert len(list(proj.iterate_components())) == 2



def test_parse_argv():
	C = fig.parse_argv(['-q', '--b', 'baba', '--a', '''{"r":10, "b":100}''',
	                    '--c', '''["asdf", 10, 4.e-4]''', '--x=99'], script_name='some-script')

	assert C.peek('a').parent is C
	assert C.root is C

	assert C.pull('b') == 'baba'
	assert C.pull('a.b') == 100
	assert C.pull('a.r') == 10
	assert C.pull('c')[0] == 'asdf'
	assert C.pull('c')[-1] == 4e-4

	assert C.peek('x').process() == 99


def test_parse_argv2():
	C = fig.parse_argv([])
	assert len(C) == 0


def test_parse_argv_meta():
	C = fig.parse_argv(['-q', '--b', 'baba', '--x=99'], script_name='some-script')

	assert C.pull('_meta.script_name') == 'some-script'
	assert C.pull('_meta.quiet')
	assert C.pull('_meta.debug', False) is False

	C = fig.parse_argv(['-q', '--b', 'baba', '--x=99', '-q'], script_name='some-script')

	assert C.pull('_meta.script_name') == 'some-script'
	assert C.pull('_meta.quiet')
	assert C.pull('_meta.debug', False) is False

	C = fig.parse_argv(['-q', '-d', '--b', 'baba', '--x=99'], script_name='some-script')

	assert C.pull('_meta.script_name') == 'some-script'
	assert C.pull('_meta.quiet')
	assert C.pull('_meta.debug')

	C = fig.parse_argv(['-q', '--b', 'baba', '--x=99', '-d'], script_name='some-script')

	assert C.pull('_meta.script_name') == 'some-script'
	assert C.pull('_meta.quiet')
	assert C.pull('_meta.debug')

	C = fig.parse_argv(['-q', '-h', '--b', 'baba', '-d', '--x=99'], script_name='some-script')

	assert C.pull('_meta.script_name') == 'some-script'
	assert C.pull('_meta.quiet')
	assert C.pull('_meta.debug')
	assert C.pull('_meta.help')



def test_config_cycle():
	reset_profile()
	fig.initialize('example7')

	try:
		fig.create_config('ec')
	except ValueError as e:
		assert str(e) == 'Config cycle detected within: ec, ec2'



def test_config_path():
	reset_profile()
	fig.initialize('example7')

	config_path = str(PROJECTS_PATH/'p7'/'config'/'e.yml')

	C = fig.create_config(config_path)

	assert C.pull('y') == 'e'


	config_path = str(PROJECTS_PATH.relative_to(Path.cwd())/'p7'/'config'/'e.yml')

	C = fig.create_config(config_path)

	assert C.pull('y') == 'e'


	C = fig.main(['get-config', config_path, 'c'])

	assert C.pull('y') == 'e'
	assert C.pull('w') == 'd'
	assert C.pull('x') == 'c'








name = 'omnifig'
long_name = 'omni-fig'

version = '1.0.7'

url = 'https://github.com/felixludos/omni-fig'

description = 'Unleashing Project Configuration and Organization'

author = 'Felix Leeb'
author_email = 'felixludos@hotmail.com'

license = 'MIT'

# readme = 'README.rst'

packages = ['omnifig']

logger_name = 'omnifig'

import os
try:
	with open(os.path.join(os.path.abspath(os.path.dirname(os.path.dirname(__file__))), 'requirements.txt'), 'r') as f:
		install_requires = f.readlines()
except:
	install_requires = ['pyyaml', 'toml', 'omnibelt>=0.8.5', 'tabulate']

entry_points = {'console_scripts': 'fig = omnifig.top:entry'}

lib_path = os.path.abspath(os.path.dirname(__file__))


__info__ = dict(locals())
for k in ['f', 'os', '__name__', '__file__', '__doc__', '__package__', '__loader__', '__spec__', '__annotations__',
          '__builtins__', '__cached__']:
	if k in __info__:
		del __info__[k]


__author__ = __info__['author']
__version__ = __info__['version']


import logging
__logger__ = logging.getLogger(__info__.get('logger_name', __name__))
del logging




from typing import Any
from ..abstract import AbstractConfig, AbstractProject
from .base import Behavior



class Quiet(Behavior, name='quiet', code='q', priority=10, num_args=0, description='Set config to silent'):
	'''
	When activated, this behavior sets the config object to silent mode,
	which means pulls/pushes are not printed to stdout
	'''
	def __init__(self, project: AbstractProject, **kwargs):
		'''Sets the attribute to keep track of what the previous value of ``config.silent`` was'''
		super().__init__(project, **kwargs)
		self.previous = None


	def pre_run(self, meta: AbstractConfig, config: AbstractConfig) -> None:
		'''
		When activated, this will set the config object to silent mode

		Args:
			meta: The meta config object (used to check if the rule is activated with the ``quiet`` key)
			config: The config object to be modified

		Returns:
			``None``

		'''
		if meta.pull('quiet', False, silent=True):
			self.previous = config.silent
			config.silent = True


	def post_run(self, meta: AbstractConfig, config: AbstractConfig, output: Any) -> None:
		'''
		When activated, this rule sets the config back to its previous value

		Args:
			meta: The meta config object (not used)
			config: Config object used to run the script.
			output: Output of the script.

		Returns:
			``None``

		'''
		if self.previous is not None:
			config.silent = self.previous


from typing import Dict, Optional, Any, Sequence, Iterator, NamedTuple, Union, ContextManager, Type
from pathlib import Path
import sys
from collections import OrderedDict
from omnibelt import Class_Registry, JSONABLE, unspecified_argument

from ..abstract import AbstractConfig, AbstractProfile, AbstractProject, AbstractBehavior
from .workspaces import ProjectBase



class ProfileBase(AbstractProfile):
	'''
	Generally, a run environment uses a single profile to keep track of loading projects,
	and invoking the top level methods (such as :func:`entry()`, :func:`main()`, :func:`run()`,
	:func:`quick_run`, etc.).

	It is recommended to subclass this class to create a custom profile classes with expected functionality
	(unlike :class:`AbstractProfile`).

	'''

	class _Behavior_Registry(Class_Registry, components=['description']):
		'''The registry for behaviors (used by projects to modify the behavior when running scripts).'''
		pass
	behavior_registry = _Behavior_Registry() # one global instance (even independent of the profile instance)


	_default_profile_cls = None
	_profile = None


	def __init_subclass__(cls, default_profile: Optional[bool] = False, **kwargs):
		super().__init_subclass__(**kwargs)
		cls._profile = None
		if default_profile is not None:
			ProfileBase._default_profile_cls = cls


	# region Class Methods
	Project: ProjectBase = ProjectBase


	@classmethod
	def get_project_type(cls, ident: str, default: Optional[Any] = unspecified_argument) -> NamedTuple:
		'''
		Gets the project type entry for the given identifier (from a registry).

		Args:
			ident: Name of the registered project type.
			default: Default value to return if the identifier is not found.

		Returns:
			Project type entry.

		'''
		return cls.Project.get_project_type(ident)


	@classmethod
	def replace_profile(cls, profile: 'ProfileBase' = None) -> 'ProfileBase':
		'''
		Replaces the current profile instance with the given profile. This is used to set the global profile.

		Args:
			profile: New profile instance.

		Returns:
			Old profile instance (which is now replaced).

		'''
		if profile is None:
			profile = cls()
		ProfileBase._profile = profile
		old = cls._profile
		cls._profile = profile
		return old


	@classmethod
	def get_profile(cls) -> 'ProfileBase':
		'''
		Gets the current profile instance of the runtime environment.

		Returns:
			Profile instance.

		'''
		if cls._profile is None:
			cls._profile = cls._default_profile_cls()
			cls._profile.activate()
		return cls._profile


	@classmethod
	def register_behavior(cls, name: str, typ: Type[AbstractBehavior], *,
	                      description: Optional[str] = None) -> NamedTuple:
		'''
		Registers a new behavior in the profile.

		Behaviors are classes which are instantiated and managed by .

		Args:
			name: Name of the behavior.
			typ: Behavior class (recommended to subclass :class:`AbstractBehavior`).
			description: Description of the behavior.

		Returns:
			Registration entry for the behavior.

		'''
		cls.behavior_registry.new(name, typ, description=description)


	@classmethod
	def get_behavior(cls, name: str) -> NamedTuple:
		'''
		Gets the behavior entry for the given identifier (from the registry).

		Args:
			name: Name of the registered behavior.

		Returns:
			Behavior entry.

		'''
		return cls.behavior_registry.find(name)


	@classmethod
	def iterate_behaviors(cls) -> Iterator[NamedTuple]:
		'''
		Iterates over all registered behaviors.

		Returns:
			Iterator over all behavior entries.

		'''
		yield from cls.behavior_registry.values()
	# endregion

	
	def __init__(self, data: Dict[str, Any] = None) -> None:
		super().__init__(data)
		self._loaded_projects = OrderedDict()
		self._current_project_key = None


	@property
	def path(self):
		path = self.data.get('info_path', None)
		if path is not None:
			return Path(path)


	# region Top Level Methods
	def entry(self, script_name: Optional[str] = None) -> None:
		'''
		Primary entry point for the profile. This method is called when using the :command:`fig` command.

		Args:
			script_name: Manually specified script name to run (if not provided, will be parsed from :code:`sys.argv`).

		Returns:
			:code:`None`

		'''
		argv = sys.argv[1:]
		self.main(argv, script_name=script_name)


	def initialize(self, *projects: str, **kwargs: Any) -> None:
		'''
		Initializes the specified projects (including activating them, which generally registers
		all associated configs and imports files and packages)

		Args:
			*projects: Identifiers of projects to initialize (activates the current project only, if none is provided).

		Returns:
			:code:`None`

		'''
		self.activate(**kwargs)
		self.get_current_project().activate()
		if len(projects):
			for project in projects:
				self.get_project(project).activate()


	def main(self, argv: Sequence[str], script_name: Optional[str] = unspecified_argument) -> None:
		'''
		Runs the script with the given arguments using :func:`main()` of the current project.

		Args:
			argv: List of top-level arguments (expected to be :code:`sys.argv[1:]`).
			script_name: specified name of the script
			(defaults to what is specified in argv when it is parsed into a config object).

		Returns:
			The output of the script.

		'''
		return self.get_current_project().main(argv, script_name=script_name)


	def run_script(self, script_name: str, config: AbstractConfig, *args: Any, **kwargs: Any) -> Any:
		'''
		Runs the script registered with the given name and the given arguments using
		:func:`run_script()` of the current project.

		Args:
			script_name: Name of the script to run (must be registered).
			config: Config object to run the script with.
			*args: Manual arguments to pass to the script.
			**kwargs: Manual keyword arguments to pass to the script.

		Returns:
			The output of the script.

		'''
		return self.get_current_project().run_script(script_name, config, *args, **kwargs)


	def run(self, config: AbstractConfig, *args: Any, **kwargs: Any):
		'''
		Runs the script with the given arguments using :func:`run()` of the current project.

		Args:
			config: Config object to run the script with.
			*args: Manual arguments to pass to the script.
			**kwargs: Manual keyword arguments to pass to the script.

		Returns:
			The output of the script.

		'''
		return self.get_current_project().run(config, *args, **kwargs)


	def quick_run(self, script_name: str, *configs: str, **parameters: Any):
		'''
		Creates a config object and runs the script using :func:`quick_run()` of the current project.

		Args:
			script_name: Name of the script to run (should be registered).
			*configs: Names of registered config files to load and merge (in order of precedence).
			**parameters: Manual config parameters to populate the config object with.

		Returns:
			Output of the script.

		'''
		return self.get_current_project().quick_run(script_name, *configs, **parameters)


	def cleanup(self, *args: Any, **kwargs: Any) -> None:
		'''
		Calls :func:`cleanup()` of the current project.

		Args:
			*args: Arguments to pass to the cleanup function.
			**kwargs: Keyword arguments to pass to the cleanup function.

		Returns:
			:code:`None`

		'''
		return self.get_current_project().cleanup(*args, **kwargs)
	# endregion


	def create_config(self, *configs: str, **parameters: JSONABLE) -> AbstractConfig:
		'''
		Process the provided data to create a config object (using the current project).

		Args:
			configs: usually a list of parent configs to be merged
			parameters: any manual parameters to include in the config object

		Returns:
			Config object resulting from loading/merging `configs` and including `data`.
		'''
		return self.get_current_project().create_config(*configs, **parameters)


	def parse_argv(self, argv: Sequence[str], script_name=None) -> AbstractConfig:
		'''
		Parses the given arguments and returns a config object.

		Arguments are expected in the following order (all of which are optional):
			1. Behaviors to modify the config loading process and script execution.
			2. Name of the script to run.
			3. Names of registered config files that should be loaded and merged (in order of precedence).
			4. Manual config parameters (usually keys, prefixed by :code:`--` and corresponding values)

		Args:
			argv: List of arguments to parse (expected to be :code:`sys.argv[1:]`).
			script_name: Manually specified name of the script (defaults to what is specified in the resulting config).

		Returns:
			Config object containing the parsed arguments.

		'''
		return self.get_current_project().parse_argv(argv, script_name=script_name)



	def __str__(self):
		return f'{self.__class__.__name__}[{self.name}]({", ".join(self._loaded_projects)})'


	def extract_info(self, other: 'ProfileBase') -> None:
		'''
		Extract data from the provided profile instance and store it in self.

		Recommended to use if a project expects a custom profile different from the currently used one.

		Args:
			profile: Base profile instance.

		Returns:
			:code:`None`

		'''
		super().extract_info(other)
		self._loaded_projects = other._loaded_projects  # .copy()
		self._current_project_key = other._current_project_key


	def get_current_project(self) -> Project:
		'''
		Gets the current project instance.

		Returns:
			Current project instance.

		'''
		return self.get_project(self._current_project_key)


	def create_project(self, name: str, path: Path = None, set_as_current: bool = True) -> AbstractProject:
		'''
		Creates a new project instance and registers it.

		Args:
			name: Name of the project.
			path: Path to the project (defaults to the current working directory).
			set_as_current: If True, sets the new project as the current project.

		Returns:
			New project instance.

		'''
		proj = self.Project(path, profile=self)
		proj.name = name
		self._loaded_projects[name] = proj
		if set_as_current:
			self._current_project_key = name
		return proj


	def switch_project(self, ident: Union[str, AbstractProject] = None) -> AbstractProject:
		'''
		Switches the current project to the one with the given identifier.

		Args:
			ident: Name of the project to switch to, defaults to the default project (with name: None).

		Returns:
			New current project instance.

		'''
		proj = self.get_project(ident)
		# current = self.get_current_project()
		# if proj is not current:
		# 	# current.deactivate()
		# 	proj.activate()
		# 	self._current_project_key = proj.name
		self._current_project_key = proj.name
		return proj


	def iterate_projects(self) -> Iterator[AbstractProject]:
		'''
		Iterates over all loaded projects.

		Returns:
			Iterator over all loaded project instances.

		'''
		yield from self._loaded_projects.values()


	def project_context(self, ident: Union[str, AbstractProject] = None) -> ContextManager[AbstractProject]:
		'''
		Context manager to temporarily switch to a different current project.

		Args:
			ident: Name of the project to switch to, defaults to the default project (with name: None).

		Returns:
			Context manager to switch to the specified project.

		'''
		return self._project_context(self, ident)


	class _project_context:
		'''
		Context manager for temporarily switching the current project.

		Args:
			ident: name or path of project to switch to

		'''

		def __init__(self, profile: AbstractProfile, ident: Union[str, Path] = None):
			self.profile = profile
			self.ident = ident
			self.old_project = None

		def __enter__(self):
			self.old_project = self.profile.get_current_project()
			self.profile.switch_project(self.ident)

		def __exit__(self, exc_type, exc_val, exc_tb):
			self.profile.switch_project(self.old_project)



def get_profile() -> ProfileBase:
	'''Returns the current profile instance.'''
	return ProfileBase.get_profile()


	





import sys, os, shutil
from omnibelt import load_export

import _test_util as tu

import omnifig as fig
# from omnifig.config import EmptyElement, ConfigType

import _test_objs
# NOTE: due to the registration when this gets imported (and the profile getting resetted in the other test scripts),
# this test must be run before any other

def test_create_obj():
	A = fig.create_config()
	A.push('a.b.c', 10)

	assert A.pull('a.b.c') == 10
	assert str(A) == 'a:\n  b: {c: 10}\n'

def test_register_load_config():
	fig.get_current_project().register_config_dir(tu.CONFIG_PATH, recursive=True)

	C = fig.create_config('test1')
	assert C.pull('arg1') == 'test'
	
	C = fig.create_config('t/h/e')
	assert C.pull('nothing') == 'special'


def test_hierarchy():
	A = fig.create_config('test2', abc=10.0)
	
	assert A.pull('a.b') == 'inside'
	assert A.pull('abc') == 10.0
	assert 'a123' not in A
	assert A.pull('fruit')[0] == 'apples'
	assert A.pull('fruit')[1] == 'bananas'
	assert 'arg1' not in A
	
	
	B = fig.create_config('test1', 'test2')
	
	assert B.pull('a.b') == 10
	assert 'a123' in B
	assert B.pull('fruit')[0] == 'tomato'
	assert B.pull('fruit')[1] == 'bananas'
	assert B.pull('arg1') == 'test'
	
	
	C = fig.parse_argv(['test1', 'test2', '--x', '''{"r":10, "z":100}''',
	                   '--arg1', 'xx',
	                   '--a.b', '11'], script_name=None)
	
	assert C.pull('a.b') == 11
	assert len(C.pull('x')) == 3
	assert C.pull('x.r') == 10
	assert C.pull('x.y') == 'hello'
	assert C.pull('x.z') == 100
	assert C.pull('arg1') == 'xx'
	
	
	D = fig.create_config('test2', fruit=["strawberries"])
	
	assert len(D.pull('fruit')) == 3
	assert D.pull('fruit')[0] == 'strawberries'
	assert D.pull('fruit')[-1] == 'peaches'


def test_deep_hierarchy():

	A = fig.create_config('test3')
	
	assert A.pull('tree') == 'nodes'
	
	assert A.pull('all')[0] == 0
	print(A.cro)
	print(type(A.cro))
	assert A.cro[0] == 'test3'
	
	order = tuple(A.cro[1:])
	assert order == ('t/n0', 't/n1', 't/n3', 't/n5', 't/n2', 't/n4', 't/n6', 't/n7')


def test_pull_simple():
	
	A = fig.create_config('test2', **{'roman.greek': 'linguistics'})
	
	assert A.pull('roman.greek') == 'linguistics'
	
	assert A.pull('simple') == 'p'
	
	assert A.pull('default', 'worked') == 'worked'
	
	assert A.pulls('not_there', 'try_again', 'alias_option') == 'found'
	
	assert A.pulls('n1', 'n2', default=2.5) == 2.5
	
	assert A.pull('unseen', silent=True) == 45

	assert A.pull('a.b') == 'inside'

	assert A.pull('x.y') == 'hello'

	a = A.peeks('x.list', 'x')
	a = a.peek('z').pull()
	assert len(a) == 2 and a[0] == 1 and a[1] == 2
	
	a = A.pulls('veggies', 'fruit')
	assert len(a) == 3 and a[0] == 'apples'
	assert type(a) == list
	
	a = A.pull('others')
	assert type(a) == list and len(a) == 3 and a[2]['not']['mickey'] == 'mouse'
	
	a = A.pull('costs')
	assert type(a) == dict and len(a) == 5 and 'rubies' in a['gems'] and a['steel'] > 1000
	
	assert A.pulls('others.2.not.mickey') == 'mouse'
	assert A.pulls('others.3', 'others.1') == 'jerry'
	assert A.pulls('others.2.not.goofy', 'fruit.0.apples', 'others.2.not')['mickey'] == 'mouse'
	
	
def test_sub():
	
	A = fig.create_config('test1')
	
	# making subs
	D = A.peek('deep')
	
	assert D.pull('x1') == 10
	assert D.pull('fruit')[0] == 'tomato'
	assert D.pull('x2.a') == 1.2
	assert D.pull('plus1') == 10
	assert D.pull('arg1') == 'test'
	
	# sub past a list
	B = A.peek('deep.qwerty.1.t')
	
	assert B.pull('d') == 2
	assert B.pull('check.1.H') == 'asdf'
	assert B.pull('x5') == 10
	assert B.pull('2.for.q')
	# assert B.pull('check.2.for.q')
	assert B.pull('difficult') == 'good'
	
	
def test_push1():
	
	A = fig.create_config('test1')
	
	assert A.pull('arg1') == 'test'
	
	a = A.push_pull('arg1', 'another')
	assert a == 'another'
	
	assert A.pull('arg1') == 'another'
	
	a = A.push_pull('a.d', 'one')
	assert a == 'one'
	assert A.pull('a.d') == 'one'
	
	a = A.push_pull('a.c.c', 123)
	assert a == 123
	assert A.pull('a.c.c') == 123


def test_push2():
	A = fig.create_config('test1')
	
	assert A.pull('fruit.0') == 'tomato'
	
	a = A.pull('fruit')
	assert len(a) == 1 and a[0] == 'tomato'
	a = A.push_pull('new_fruit', a)
	assert len(a) == 1 and a[0] == 'tomato'
	
	a = A.push_pull('new_fruit', 'no')
	assert a == 'no'
	assert A.pull('new_fruit.0', 'failed') == 'failed'
	assert A.pull('new_fruit') == 'no'
	
	assert A.pull('a.b', 10)
	a = A.push_pull('a.b', 11)
	assert a == 11
	assert A.pull('a.b') == 11
	# a = A.push('a.b.c', 0.25)
	# assert a == 0.25
	# assert A.pull('a.b')['c'] == 0.25
	
	a = A.push_pull('a123', 'removed')
	assert a == 'removed'
	
	a = A.push_pull('new', [1,2,4])
	assert a[-1] == 4
	assert A.pull('new.0') == 1
	
	a = A.push_pull('nd', ['no', {'k':[0]}])
	assert a[0] == 'no'
	assert 'k' in a[1]
	a = A.pull('nd')
	assert len(a) == 2 and a[0] == 'no' and a[1]['k'][0] == 0
	assert A.pull('nd.0') == 'no'
	assert A.pull('nd.o', 'w') == 'w'
	assert A.pulls('nd.1.j', 'nd.0') == 'no'
	assert A.pulls('deep.x1.qwer.asdf', 'nd.1.k.0') == 0


def test_push3():
	A = fig.create_config('test1')
	
	# child pushes
	a = A.push_pull('arg2.a.b', [-1,10])
	assert a[0] == -1
	assert A.pull('arg2.a')['b'][1] == 10
	assert A.pull('arg2.a.b.1') == 10
	
	# when parent exists
	a = A.push_pull('deep.unknown', 'gooder')
	assert a == 'gooder'
	assert A.pull('unknown') == 'bad'
	assert A.pull('deep.qwerty.1.t.unknown') == 'good'
	assert A.pull('deep.unknown') == 'gooder'
	
	# push to list (append and indexed)
	a = A.push_pull('fruit.1', 'yummy')
	assert a == 'yummy'
	assert A.pull('fruit.1') == 'yummy'
	a = A.push_pull('fruit.2', 'blueberry')
	assert a == 'blueberry'
	assert len(A.pull('fruit')) == 3
	assert A.pull('fruit.2') == 'blueberry'
	
	a = A.push_pull('fruit.10', 'veggies')
	assert a == 'veggies'
	assert A.pull('fruit.3') == A.empty_value
	assert A.pull('fruit.7') == A.empty_value
	assert A.pull('fruit.10') == 'veggies'
	assert A.pull('fruit.11', 'nope') == 'nope'
	assert A.pull('fruit.20', 'no way') == 'no way'


def test_push4():
	A = fig.create_config('test1')
	
	a = A.push_pull('test', 'not me')
	assert a == 'not me'
	assert A.pull('deep.x1') == 10
	assert A.pull('deep')['x1'] == 10
	
	a = A.push_pull('deep.qwerty.0', {'x1': 11})
	assert a['x1'] == 11
	
	assert A.pull('deep.qwerty.0')['x1'] == 11
	assert A.pull('deep.qwerty.0.x1') == 11
	assert A.pulls('deep.qwerty.1.x1', 'deep.x1') == 10
	
	assert A.pull('deep.x1') == 10
	assert A.pull('deep')['x1'] == 10
	# assert A.pull('deep.qwerty.x1') == 10
	
	a = A.push_pull('deep.qwerty.2', 'no more')
	assert a == 'no more'
	assert A.pull('deep.qwerty.2') == 'no more'
	
	a = A.push_pull('deep.qwerty.3', [100])
	assert len(a) == 1 and a[0] == 100
	assert A.pull('deep.qwerty.3.0') == 100
	
	# with/out overwriting
	a = A.push_pull('deep.qwerty.4', 23, overwrite=False)
	assert a == 23
	assert A.pull('deep.qwerty.4') == 23
	
	a = A.push_pull('a.x', 2, overwrite=False)
	assert a == 1
	assert A.pull('a.x') == 1


def test_create_silent():
	import io, contextlib

	buffer = io.StringIO()
	with contextlib.redirect_stdout(buffer):
		A = fig.create_config(arg1='test')

		node = A.peek('arg1')
		assert node.create_silent() == 'test'

	printed = buffer.getvalue()
	assert printed == ''



def test_aliases():
	
	A = fig.create_config('test4')
	B = A.peek('l4.l5')
	
	assert A.pull('s1') == 'f1'
	assert B.pull('s2') == 'f2'


def test_update_sub():
	
	A = fig.create_config('test1')
	
	# update with config (check parents)
	
	B = fig.create_config()
	B.push('a', {'x':10, 'c': 20})
	B.push('fruit', 'not a list')
	B.push('a123', ['pears', 'raspberries'])

	assert B.pull('a123.0') == 'pears'
	assert A.pull('a123.0') == 1
	
	assert A.pull('a.x') == 1
	assert B.pull('a.x') == 10
	
	B.push('a.d.a', 'micro')
	assert B.pull('a.d.a')
	
	B.push('q', {'y': {'m': {'c': {'a': 'hey'}}}})
	B.push('alpha', [-9,2.1,{'beta':[-4,{'gamma'}]}])

	b = B.pull('alpha.2.beta.1')
	assert type(b) is set and 'gamma' in b
	
	A.update(B)
	
	assert A.pull('a.x') == 10
	assert A.pull('q.y.m.c.a') == 'hey'
	assert A.pull('fruit') == 'not a list'
	assert A.pull('a123.0') == 'pears'
	assert A.pull('a123.1') == 'raspberries'
	assert A.pull('a123.2') == 'another'
	assert 'gamma' in B.pull('alpha.2.beta.1')
	
	C = A.peek('q.y.m.c')
	
	assert C.pull('a') == 'hey'
	assert C.pull('unknown') == 'bad'
	assert C.pull('c.a') == 'hey'
	
	c = C.push_pull('m', 'tick')
	assert c == 'tick'
	assert C.pull('m') == 'tick'
	assert C.pull('y.m')['c']['a'] == 'hey'
	
	D = A.peek('alpha.2.beta')
	
	assert D.pull('0') == -4
	assert D.pull('alpha.0') == -9
	assert D.pull('unknown') == 'bad'


def test_repeat_create():
	raw = ['tomato']
	A = fig.create_config(fruit=raw)

	node = A.peek('fruit')
	fruit = node.create()

	assert raw is not fruit
	assert isinstance(fruit, list) and len(fruit) == 1 and fruit[0] == 'tomato'

	cached = node.pull()
	assert cached is not fruit
	cached2 = node.process()
	assert cached is cached2
	cached3 = A.pull('fruit')
	assert cached is cached3

	fruit2 = node.create()
	assert fruit2 is not fruit
	assert fruit2 == fruit

	fruit3 = node.create()
	assert fruit3 is not fruit2
	assert fruit3 == fruit2

	
def test_update_dict():

	A = fig.create_config('test1')
	
	assert A.pull('a.x') == 1
	
	A.update(A.from_raw({'y': {'m': {'c': {'a': 'hey'}}}, 'a': {'x':12}}))
	
	assert A.pull('a.x') == 12
	assert A.pull('y.m.c.a') == 'hey'
	
	C = A.peek('y.m.c')
	
	assert C.pull('a') == 'hey'
	assert C.pull('unknown') == 'bad'
	assert C.pull('c.a') == 'hey'
	
	A.update(A.from_raw({'alpha':[-9,2.1,{'beta':[-4,{'gamma'}]}]}))
	
	assert A.pull('alpha.0') == -9
	assert tuple(A.pull('alpha.2.beta.1')) == ('gamma',)
	
	D = A.peek('alpha.2.beta')
	
	assert D.pull('0') == -4
	assert D.pull('alpha.0') == -9
	assert D.pull('unknown') == 'bad'


def test_export():
	fig.get_current_project().register_config_dir(tu.CONFIG_PATH, recursive=True)

	# load/change/export
	A = fig.create_config('test2', **{'a.d': 50})

	A.push('count', [40,40])

	# path = os.path.join(tu.TEST_PATH, 'save.yaml')
	root = tu.TEST_PATH / 'temp-exports'
	root.mkdir(exist_ok=True)
	path = root / 'save.yaml'
	A.export(path)

	# reload from path (rel and abs) check for change

	B = fig.create_config(path)

	assert B.pull('a.b') == 'inside'
	assert B.pull('unseen') == 45
	assert B.pull('count.0') == 40
	assert B.pull('a.d') == 50
	assert B.pull('others.2.not.mickey') == 'mouse'

	B.push('a.d', 100)

	path = B.export('save2', root=root)

	C = load_export(path)

	assert C.pull('a.b') == 'inside'
	assert C.pull('unseen') == 45
	assert C.pull('count.0') == 40
	assert C.pull('a.d') == 100
	assert C.pull('others.2.not.mickey') == 'mouse'

	if root.exists():
		shutil.rmtree(root)
	

def test_components():
	A = fig.create_config('test5')
	
	# create
	c = A.peek('a7').create()
	
	assert c.a2 == 'worked'
	assert c.a4_a6 == 2
	assert c.a10_0 == '0'
	assert c.x is None
	assert c.a14 == 'success'
	assert c.a14b == 'next'
	
	assert c.f(10) == 11
	
	# create nested
	A.push('a7.mn._type', 'c2')
	A.push('a7.mn.a3', 'billion')
	
	d = A.pull('a7')
	
	assert d.x is not None
	assert d.x.a1 == 1
	assert d.x.a2 == 'worked'
	assert d.x.a3 == 'billion'
	
	d.delta = 1234
	
	# reuse
	e = A.pull('a7')
	
	assert id(e) == id(d)
	assert id(d.x) == id(e.x)
	assert e.delta == d.delta
	assert e.delta == 1234
	
def test_modifiers():
	
	A = fig.create_config('test6', 'test5')
	
	# simple auto modifiers
	
	c = A.pull('a7')
	
	assert type(c).__name__ == 'M2_M1_C1'
	assert c.f(1) == 500
	assert c.g(2) == 998
	assert c.h() == 502
	assert c.x is None
	
	# multiple auto modifiers
	
	A.push('a7._mod', 'm1')
	
	d = A.peek('a7').create()
	
	assert type(d).__name__ == 'M1_C1'
	assert d.f(1) == 0
	assert d.g(10) == -11
	assert d.h() == 2
	
	
	# modifications
	
	# A.push('a7._mod', 'm3')
	#
	# e = A.peek('a7').create()
	#
	# assert type(e).__name__ == 'C1'
	# assert e.a == 'still worked'
	
	
def test_iteration():
	
	A = fig.create_config('test2')
	
	# iterate through a list with as_iter
	
	# itr = A.pull('others', as_iter=True)
	node = A.peek('others')
	itr = node.pull_children()
	
	assert len(node) == 3
	assert next(itr) == 'tom'
	# assert len(itr) == 2
	assert next(itr) == 'jerry'
	assert next(itr)['not']['mickey'] == 'mouse'
	
	# iterate through a dict with as_iter
	
	A.push('costs.gems', '_x_')
	
	answers = A.peek('costs').create()

	node = A.peek('costs')
	itr = node.pull_named_children()
	
	assert len(node) == 4
	for _ in range(len(node)):
		k,v = next(itr)
		
		assert k in answers
		assert answers[k] == v
	
	# auto iterate a list with seq
	
	node = A.peek('others')
	itr = node.peek_children()
	
	assert len(node) == 3
	assert next(itr).create() == 'tom'
	# assert len(itr) == 2
	# assert next(itr).create() == 'jerry'
	assert next(itr).parent == node
	assert next(itr).create()['not']['mickey'] == 'mouse'
	

	# auto iterate a dict with seq
	
	node = A.peek('costs')
	itr = node.peek_named_children()
	
	assert len(node) == 4
	for _ in range(len(node)):
		k, v = next(itr)
		
		assert k in answers
		assert v.parent == node
		assert answers[k] == v.create()
	
	
def test_removing():
	
	A = fig.create_config('test2')
	
	assert A.pull('papageno') == 10
	assert A.pull('papagena', 'missing') == 'missing'
	
	A.push('vogel.papagena', '_x_')
	
	node = A.peek('vogel')
	itr = node.pull_named_children()
	
	assert len(node) == 1
	assert tuple(next(itr)) == ('papagei', 'parrot')
	

def test_raw_and_cousins():
	
	A = fig.create_config('test5')
	
	raw = A.peek('a7')
	assert isinstance(raw, type(A))
	
	A.push('pocket.a7', raw)
	A.push('a2', '_x_')
	A.push('a7', '_x_')
	A.push('a7.a2', 'cousin')
	A.push('pocket.a7.a2', '_x_')
	
	A.settings['allow_cousins'] = True
	
	pocket = A.peek('pocket')
	
	# create
	c = pocket.pull('a7')
	
	assert c.a2 == 'cousin'
	assert c.a4_a6 == 2
	assert c.a10_0 == '0'
	assert c.x is None
	assert c.a14 == 'success'
	assert c.a14b == 'next'

	# testing printing

	print(A)
	print()
	print(fig.create_config('test3'))
	
	
# TODO: unit tests checking the printed output when pulling/pushing objects


def test_pull_cycle():

	A = fig.create_config()

	A.push('a', ['<>a'])

	try:
		out = A.pull('a')
	except A.CycleError as e:
		assert e.config.my_address() == ('a',)
	else:
		assert False, 'CycleError not raised'


def test_underscores():

	# NOTE: dashes default to underscores (but not vice versa)
	# so, in general, queries should use dashes, configs should use underscores

	A = fig.create_config(a_b=1, _b=2, **{'-c': 3, 'd-e': 4, '--f': 5})

	assert A.pull('a_b') == 1
	assert A.pull('_b') == 2
	# assert A.pull('_c') == 3
	# assert A.pull('d_e') == 4
	# assert A.pull('__f') == 5

	assert A.pull('a-b') == 1
	assert A.pull('-b') == 2
	assert A.pull('-c') == 3
	assert A.pull('d-e') == 4
	assert A.pull('--f') == 5

	print()
	print(A)


def test_underscores2():
	data = {'-x':1, '_y': '-a_'}
	A = fig.create_config(a_b=data, _b=2, **{'-c': data, 'd-e': 4, '--f': 5})

	assert A.pull('a_b') == data
	assert A.pull('a-b') == data
	assert A.pull('-c') == data

	assert A.pull('a_b.-x') == 1
	assert A.pull('a_b._y') == '-a_'

	assert A.pull('a-b.-x') == 1
	assert A.pull('a-b._y') == '-a_'

	assert A.pull('-c.-x') == 1
	assert A.pull('-c.-y') == '-a_'












import sys
import omnifig as fig

@fig.script('print-config', description='Prints out the config (for debugging)')
def print_config(config): # just prints out the config
	'''
	Here's the doc string of the "config" script

	Args:
		config: object created by omni-fig

	Returns:
		the config object

	'''

	print('This is the config:')
	print(config)

	return config


@fig.autocomponent('stdout')  # automatically pulls all arguments in signature before creating
def _get_stdout():  # in this case, we don't need any arguments
	return sys.stdout


@fig.autocomponent('file')
def _get_file(path):
	return open(path, 'a+')







from typing import List, Dict, Tuple, Optional, Union, Any, Sequence, Type, Iterator, NamedTuple, ContextManager
from pathlib import Path
from contextlib import nullcontext
from copy import deepcopy
import yaml
from collections import OrderedDict
from omnibelt import unspecified_argument, Primitive, primitive, Modifiable
from omnibelt.nodes import AutoTreeNode, AutoTreeSparseNode, AutoTreeDenseNode

from .. import __logger__ as prt
from ..abstract import AbstractConfig, AbstractProject, AbstractCreator, AbstractConfigurable, AbstractConfigManager, \
	AbstractCustomArtifact, AbstractCertifiable

from .abstract import AbstractSearch, AbstractReporter



class ConfigNode(AutoTreeNode, AbstractConfig):
	'''
	The main config node class. This class is used to represent the config tree
	and is the main interface for interacting with the config.
	'''

	Settings = OrderedDict

	@classmethod
	def from_raw(cls, raw: Any, *, parent: Optional['ConfigNode'] = unspecified_argument,
	             parent_key: Optional[str] = None, **kwargs) -> 'ConfigNode':
		'''
		Converts the given raw data into a config node.
		This will recursively convert all nested data into config nodes.

		Args:
			raw: python data to convert (may be a primitive or a dict/list-like object)
			parent: the parent node (if any) of the node to be created
			parent_key: the key of the node to be created (if any) in the parent node
			**kwargs: additional arguments to pass to the constructor

		Returns:
			The created config node

		'''
		if isinstance(raw, ConfigNode):
			raw.parent = parent
			raw._parent_key = parent_key
			return raw
		if isinstance(raw, dict):
			node = cls.SparseNode(parent=parent, parent_key=parent_key, **kwargs)
			for key, value in raw.items():
				child = cls.from_raw(value, parent=node, parent_key=key, **kwargs)
				if key in node:
					node.get(key).update(child)
				else:
					node.set(key, child, **kwargs)
		elif isinstance(raw, (tuple, list)):
			node = cls.DenseNode(parent=parent, parent_key=parent_key, **kwargs)
			for idx, value in enumerate(raw):
				idx = str(idx)
				node.set(idx, cls.from_raw(value, parent=node, parent_key=idx, **kwargs), **kwargs)
		else:
			node = cls.DefaultNode(payload=raw, parent=parent, parent_key=parent_key, **kwargs)
		return node
	
	
	class Search(AbstractSearch):
		'''
		Used to traverse the config tree and find the node corresponding to the given set of queries.
		'''

		confidential_prefix = '_' # keys with this prefix do not default to parent nodes

		force_create_prefix = '<!>' # force create of delegate
		delegation_prefix = '<>' # delegate from current node
		delegation_origin_prefix = '<o>' # delegate from origin node of the query
		missing_key_payload = '__x__' # payload for a node that should be treated as missing

		class _sub_search:
			'''
			Context manager to keep track of the previous search object when the search is nested
			(e.g. when the product is a container such as a list or dict).
			'''
			past = None

			def __init__(self, current: AbstractSearch):
				self._old = self.past
				self.current = current
				self.__class__.past = current

			def __enter__(self):
				pass

			def __exit__(self, exc_type, exc_val, exc_tb):
				self.__class__.past = self._old


		def sub_search(self) -> 'ContextManager':
			'''Returns a context manager that allows new searches to reference back to this search'''
			return self._sub_search(self)


		def __init__(self, origin: 'ConfigNode', queries: Optional[Sequence[str]], default: Any,
		             parent_search: Optional['ConfigNode.Search'] = unspecified_argument, **kwargs):
			'''
			Evaluates the given queries and default values to find the node (or consequent product) in the config tree.

			Args:
				origin: the node from which to start the search
				queries: keys to search for (in order) in the config tree
				default: if the search fails, this value is returned
				parent_search: if the search is nested, this is the search that was used to find the parent node
				**kwargs: additional arguments to pass to the constructor (not used)
			'''
			if parent_search is unspecified_argument:
				parent_search = self._sub_search.past
			super().__init__(origin=origin, queries=queries, default=default, **kwargs)
			self.origin = origin
			self.queries = queries
			self.default = default
			self.query_chain = []
			self.result_node = None
			self.force_create = False
			self.parent_search = parent_search
			self.extra_queries = None


		def _resolve_query(self, src: 'ConfigNode', query: str, *remaining: str,
		                   chain: Optional[List[str]] = None) -> Tuple[Optional['ConfigNode'], Tuple, List[str]]:
			'''
			Traverses the config tree from the given source node to find the node corresponding to the given query,
			thereby appending the query to the query chain.

			Note that when a query is not found directly in the src node (i.e. the query is not a key in the src node).
			If that fails and the current node has a parent, and the query does not begin with "_" the query is
			passed to the parent node as is. If that also fails and the "allow_cousins" setting is on,
			the parent_key is prepended to the query and passed to the grandparent node (if one exists).

			This process relies on two flags in the settings of the config: "ask_parent" (default: True)
			and "allow_cousins" (default: False).

			Args:
				src: the current node in the traversal
				query: the target key to search for
				*remaining: alternative queries to search for (if the current query fails)
				chain: chain of queries that have been traversed so far (e.g. due to delegations or missing keys)

			Returns:
				A tuple of the node when resolving all given queries in order (or None if no node was found),
				any remaining unused queries, and the updated query chain

			'''
			if chain is None:
				chain = []
			if query is None:
				return src, remaining, chain

			try:
				result = src.get(query)
			except src._MissingKey:
				result = None

			if result is None:
				if src.settings.get('ask_parents', True) and not query.startswith(self.confidential_prefix):
					parent = src.parent
					if parent is not None:
						result, _, parent_chain = self._resolve_query(parent, query)
						if result is None:
							grandparent = parent.parent
							if grandparent is not None and src.settings.get('allow_cousins', False) \
									and src._parent_key is not None:
								cousin_query = f'{src._parent_key}.{query}'
								result, _, cousin_chain = self._resolve_query(grandparent, cousin_query)
								if result is not None:
									chain.append(f'..{cousin_chain[-1]}')
									return result, remaining, chain
						else:
							chain.append(f'.{parent_chain[-1]}')
							return result, remaining, chain
				if len(remaining):
					return self._resolve_query(src, *remaining, chain=chain)
			chain.append(query)
			return result, remaining, chain


		def _find_node(self) -> 'ConfigNode':
			'''
			Traverses the config tree from the origin node to find the node corresponding to the queries
			(given in __init__). If no queries are provided, the origin node is returned.

			Before returning the node, the node is processed with the "process_node"
			(which should check follow delegations as needed). The output of "process_node" is stored as "result_node",
			while the original result of resolving the queries without processing is stored as "query_node".
			Lastly, any unused queries are stored as "unused_queries", while the query chain thus far is "query_chain".

			Returns:
				The processed node corresponding to the queries

			Raises:
				SearchFailed: if the node could not be found

			'''
			if self.queries is None or not len(self.queries):
				result = self.origin
			else:
				result, self.unused_queries, self.query_chain = self._resolve_query(self.origin, *self.queries)
			self.query_node = result
			self.result_node = self.process_node(result)
			return self.result_node


		def find_node(self, silent: Optional[bool] = None) -> 'ConfigNode':
			'''
			Traverses the config tree from the origin node to find the node corresponding to the queries
			(given in __init__). If no queries are provided, the origin node is returned.

			This method should be used when the end result of the search should be the node (not a product).

			Returns:
				The node corresponding to the queries with the current search as the trace (for reporting)

			Raises:
				SearchFailed: if the node could not be found and no default was provided

			'''
			if silent is None:
				silent = self.origin.silent
			try:
				result = self._find_node()
			except self.SearchFailed:
				if self.default is self.origin._empty_default:
					raise
				return self.default
			result._trace = self
			self.origin.reporter.report_node(result, silent=silent)
			return result


		def find_product(self, silent: Optional[bool] = None) -> Any:
			'''
			Traverses the config tree from the origin node to find the node corresponding to the queries and then
			produces the product of that node (using the "process" or "create" method of the node).

			This method should be used when the end result of the search should be the product
			(i.e. value contained by the node).

			If all the queries failed, the default is returned (if provided) and reported using the origin's reporter.

			Args:
				silent: propagate the silent flag the reporter or node processing

			Returns:
				Result of resolving the queries, which is either the product of the node
				or the default value if no node was found

			Raises:
				SearchFailed: if the node could not be found and no default was provided

			'''
			if silent is None:
				silent = self.origin.silent
			try:
				node = self._find_node()
			except self.SearchFailed:
				if self.default is self.origin._empty_default:
					raise
				old = self.origin._trace
				self.origin._trace = self
				self.origin.reporter.report_default(self.origin, self.default, silent=silent)
				self.origin._trace = old
				result = self.default
			else:
				if node is self.origin.empty_value:
					result = node # TODO: finish
					old = self.origin._trace
					self.origin._trace = self
					self.origin.reporter.report_empty(self.origin, silent=silent)
					self.origin._trace = old
				else:
					old = node._trace
					node._trace = self
					result = node._create(silent=silent) if self.force_create \
						else node._process(silent=silent)
					node._trace = old
			return result


		def process_node(self, node: 'ConfigNode') -> 'ConfigNode':  # resolves delegations
			'''
			Processes the node by checking for delegations (and then resolving those)
			or any other special search features (such as making sure this node is still valid).

			A delegation is a node where the value is itself interpretted as a query to a different node
			(e.g. using the prefix "<>"). There are a few specific prefixes that can be used for different types of
			delegations:
				
				- ``<>``: delegate to a new node starting the search from the current node
				
				- ``<o>``: delegate to a new node starting the search from the origin node
				
				- ``<!>``: delegate to a new node for which the product is always newly created
					(rather than reused if it already exists)

			If the current node delegates, the unused queries and query chain may be updated through
			the ``_resolve_query()`` method.

			If the value of a node is "__x__", the node is considered to be missing
			(i.e. it effectively doesn't exist).

			Args:
				node: the result of resolving the queries thus far

			Returns:
				The node once all delegations and special features have been resolved

			Raises:
				SearchFailed: if the node is not valid or a delegation could not be resolved

			'''
			if node is None:
				raise self.SearchFailed(*self.query_chain)
			if self.origin.empty_value is node:
				return node
			if node.has_payload:
				payload = node.payload
				if isinstance(payload, str):
					if payload.startswith(self.delegation_prefix):
						ref = payload[len(self.delegation_prefix):]
						result, self.unused_queries, self.query_chain \
							= self._resolve_query(node, ref, *self.unused_queries, chain=self.query_chain)
						return self.process_node(result)
					elif payload.startswith(self.delegation_origin_prefix):
						ref = payload[len(self.delegation_origin_prefix):]
						result, self.unused_queries, self.query_chain \
							= self._resolve_query(self.origin, ref, *self.unused_queries, chain=self.query_chain)
						return self.process_node(result)
					elif payload.startswith(self.force_create_prefix):
						ref = payload[len(self.force_create_prefix):]
						result, self.unused_queries, self.query_chain \
							= self._resolve_query(node, ref, *self.unused_queries, chain=self.query_chain)
						self.force_create = True
						return self.process_node(result)
					elif payload == self.missing_key_payload:
						result, self.unused_queries, self.query_chain \
							= self._resolve_query(self.query_node, *self.unused_queries, chain=self.query_chain)
						return self.process_node(result)
			return node

	SearchFailed = Search.SearchFailed


	class Reporter(AbstractReporter):
		'''Formats and prints the results of a search over the config tree.'''
		def __init__(self, indent: str = ' > ', flair: str = '| ', alias_fmt: str = ' --> ', colon:str = ': ',
		             max_num_aliases: int = 3, **kwargs):
			'''
			Specifies the format in which the search results are printed to the console.

			Args:
				indent: for each depth, this string is prepended to the line
				flair: prepended to every line this reporter prints (to distinguish it from other output)
				alias_fmt: used to indicate that a query was replaced by another (e.g. due to delegation or missing)
				colon: separates the key from the value
				max_num_aliases: the maximum number of aliases to print before truncating the list
				**kwargs: passed to the parent class (unused)
			'''
			super().__init__(**kwargs)
			self.indent = indent
			self.flair = flair
			self.alias_fmt = alias_fmt
			self.colon = colon
			self.max_num_aliases = max_num_aliases


		@classmethod
		def _node_depth(cls, node: 'ConfigNode', _fuel: int = 1000) -> int:
			'''
			Returns the depth of the node in the config tree for indents.

			Args:
				node: the node to check
				_fuel: used to prevent infinite recursion

			Returns:
				The depth of the node in the config tree

			Raises:
				RecursionError: if the node is too deep in the tree
			'''
			if _fuel <= 0:
				raise RecursionError('Depth exceeded 1000 (there is probably an infinite loop in the config tree)')
			if node.parent is None:
				return 0
			return cls._node_depth(node.parent, _fuel=_fuel - 1) + 1


		@staticmethod
		def log(*msg: str, end: str = '\n', sep: str = ' ', silent: Optional[bool] = None) -> str:
			'''
			Prints the message to the console (similar to :code:`print`).

			Args:
				*msg: terms to join and print
				end: ending character (default is newline)
				sep: character to join terms (default is space)
				silent: if True, the message is not printed (but still returned)

			Returns:
				The message that was printed

			'''
			msg = sep.join(str(m) for m in msg) + end
			if not silent:
				print(msg, end='')
			return msg


		def get_key(self, trace: 'ConfigNode.Search') -> str:
			'''
			Formats the key of the node (taking parents into account).

			Args:
				trace: the search context (used to get the node)

			Returns:
				The formatted key (including aliases and parents)

			'''
			if trace is None:
				return '.'
			queries = trace.query_chain

			if trace.parent_search is not None:
			# if len(search.parent_search):
				queries = queries.copy()
				queries[0] = f'({queries[0]})'  # if getattr(search.parent_search[0]

			if len(queries) > self.max_num_aliases:
				key = self.alias_fmt.join([queries[0], '...', queries[-1]])
			else:
				key = self.alias_fmt.join(queries)
			return key.replace('_', '-')


		def _stylize(self, node: 'ConfigNode', line: str) -> str:
			'''
			Stylizes the line based on the search origin's depth and the reporter's flair.

			Args:
				node: result of the search
				line: information to print

			Returns:
				The stylized line (including flair and indents)

			'''
			trace = node.trace
			if trace is not None:
				node = trace.origin
			indent = self._node_depth(node) * self.indent
			return f'{self.flair}{indent}{line}'


		def _format_component(self, key: str, component_type: str, modifiers: Sequence[str],
		                      creator_type: Optional[str]) -> str:
			'''
			Formats the component information (type, modifiers, creator) when it is created.

			Args:
				key: of the node that contains the component (i.e. result of :code:`get_key`)
				component_type: registered type of the component
				modifiers: registered modifiers of the component
				creator_type: registered type of the creator (defaults to None)

			Returns:
				The formatted component information

			'''
			mods = modifiers
			mod_info = ''
			if len(mods):
				mod_info = f' (mods=[{", ".join(map(repr,mods))}])' if len(mods) > 1 else f' (mod={mods[0]!r})'
			if creator_type is not None:
				mod_info = f'{mod_info} (creator={creator_type!r})'
			key = '' if key is None else f'{key} '
			return f'{key}type={component_type!r}{mod_info}'


		def _format_value(self, value: Any) -> str:
			'''
			Formats the value of the node (e.g. when the product is a python primitive).

			Args:
				value: payload of the node

			Returns:
				The formatted value (using :code:`repr`)

			'''
			return repr(value)


		def report_node(self, node: 'ConfigNode', *, silent: bool = None) -> Optional[str]:
			'''
			Reports when a node was found and prints it to the console. By default, no message is printed.

			Args:
				node: result of the search
				silent: if True, the message is not printed (but still returned)

			Returns:
				None

			'''
			pass


		def report_default(self, node: 'ConfigNode', default: Any, *, silent: bool = None) -> Optional[str]:
			'''
			Reports when a default value was used and prints it to the console.

			Args:
				node: result of the search
				default: value that was used instead
				silent: if True, the message is not printed (but still returned)

			Returns:
				Message that was printed

			'''
			trace = node.trace
			key = self.get_key(trace)

			line = f'{key}{self.colon}{self._format_value(default)} (by default)'
			return self.log(self._stylize(node, line), silent=silent)


		def report_empty(self, node: 'ConfigNode', *, silent: bool = None) -> Optional[str]:
			'''
			Reports when a node was found, but it was empty and prints it to the console.

			Args:
				node: result of the search
				silent: if True, the message is not printed (but still returned)

			Returns:
				Message that was printed

			'''
			trace = node.trace
			key = self.get_key(trace)

			line = f'{key}{self.colon}<is empty>'
			return self.log(self._stylize(node, line), silent=silent)


		def report_iterator(self, node: 'ConfigNode', product: Optional[bool] = False, *,
		                    silent: Optional[bool] = None) -> Optional[str]:
			'''
			Reports when a node was found and returned as an iterator and prints it to the console.

			Args:
				node: result of the search
				product: if True, the iterator returns the products of the nodes (defaults to False)
				silent: if True, the message is not printed (but still returned)

			Returns:
				Message that was printed

			'''
			trace = node.trace
			key = self.get_key(trace)
			N = len(node)
			size = f' [{N} element{"s" if N == 0 or N > 1 else ""}]'
			return self.log(self._stylize(node, f'ITERATOR {key}{size}'), silent=silent)


		def reuse_product(self, node: 'ConfigNode', product: Any, *, silent: bool = None) -> Optional[str]:
			'''
			Reports when a node was found and its product was reused and prints it to the console.

			Args:
				node: result of the search
				product: the product that was reused
				silent: if True, the message is not printed (but still returned)

			Returns:
				Message that was printed

			'''
			trace = node.trace
			key = self.get_key(trace)

			reusing = '' if isinstance(product, primitive) else ' (reuse)'
			line = f'{key}{self.colon}{self._format_value(product)}{reusing}'
			return self.log(self._stylize(node, line), silent=silent)

			line = f'REUSING {self._format_component(key, component_type, modifiers, creator_type)}'
			return self.log(self._stylize(node, line), silent=silent)


		def create_primitive(self, node: 'ConfigNode', value: Primitive = unspecified_argument, *,
		                     silent: bool = None) -> Optional[str]:
			'''
			Reports when a product was created that was a primitive and prints it to the console.

			Args:
				node: result of the search
				value: of the product
				silent: if True, the message is not printed (but still returned)

			Returns:
				Message that was printed

			'''
			trace = node.trace
			key = self.get_key(trace)

			if value is unspecified_argument:
				value = node.payload
			line = f'{key}{self.colon}{self._format_value(value)}'
			return self.log(self._stylize(node, line), silent=silent)


		def create_container(self, node: 'ConfigNode', *, silent: bool = None) -> Optional[str]:
			'''
			Reports when a product was created that was a container and prints it to the console.

			Args:
				node: result of the search
				silent: if True, the message is not printed (but still returned)

			Returns:
				Message that was printed

			'''
			trace = node.trace
			if trace is not None:
				key = self.get_key(trace)
				N = len(node)

				t, x = ('dict', 'item') if isinstance(node, trace.origin.SparseNode) else ('list', 'element')
				x = f'{x}s' if N != 1 else x
				line = f'{key} [{t} with {N} {x}]'
				return self.log(self._stylize(node, line), silent=silent)


		def create_component(self, node: 'ConfigNode', *, component_type: str = None,
		                     modifiers: Optional[Sequence[str]] = None, creator_type: str = None,
		                     silent: bool = None) -> Optional[str]:
			'''
			Reports when a product was created that was a component and prints it to the console.

			Args:
				node: result of the search
				component_type: registered name of the component
				modifiers: registered names of the modifiers
				creator_type: registered name of the creator (defaults to None)
				silent: if True, the message is not printed (but still returned)

			Returns:
				Message that was printed

			'''
			trace = node.trace
			key = self.get_key(trace)
			line = f'CREATING {self._format_component(key, component_type, modifiers, creator_type)}'
			return self.log(self._stylize(node, line), silent=silent)


	class CycleError(RuntimeError):
		'''Raised when a cycle is detected in the config tree.'''
		def __init__(self, config):
			super().__init__(f'Cycle detected for {config.my_address()}')
			self.config = config


	class DefaultCreator(AbstractCreator):
		'''
		Manages the creation of products of config nodes. Generally, there are three types of products:
			- primitives (int, float, str, bool, None)
			- containers (list, dict)
			- components (objects, must be registered, and may optionally be modified)

		The default creator is responsible for creating the products, but you can also create
		your own creators (e.g. subclassses) and then specify them when registering components
		(to make those the defaults) or in the config directly.
		Alternatively, you can force a specific creator to be used by changing the config setting "creator".

		Note, that it is generally up to the creator to call the config reporter to report the creation of products.

		'''
		_config_component_key = '_type'
		_config_modifier_key = '_mod'
		_config_creator_key = '_creator'

		_creation_context = None
		
		@classmethod
		def replace(cls, creator: 'ConfigNode.DefaultCreator', config, *, component_type: Optional[str] = None,
		            modifiers: Optional[Sequence[str]] = None,
		            project: Optional[AbstractProject] = unspecified_argument,
		            component_entry=unspecified_argument, silent=unspecified_argument, **kwargs) -> AbstractCreator:
			'''
			Extracts information from the given creator to replace it.
			Used primarily in :meth:`DefaultCreator.validate`.
			'''
			if component_type is None:
				component_type = creator.component_type
			if modifiers is None:
				modifiers = creator.modifiers
			if project is unspecified_argument:
				project = creator.project
			if component_entry is unspecified_argument:
				component_entry = creator.component_entry
			if silent is unspecified_argument:
				silent = creator.silent
			return super().replace(creator, config, component_type=component_type, modifiers=modifiers, project=project,
			                       component_entry=component_entry, silent=silent, **kwargs)


		def __init__(self, config: 'ConfigNode', *, component_type: Optional[str] = unspecified_argument,
		             modifiers: Optional[Sequence[str]] = None, project: Optional[AbstractProject] = None,
		             component_entry: Optional[NamedTuple] = unspecified_argument, silent: Optional[bool] = None,
		             **kwargs):
			'''
			A default creator is instantiated each time a product of a config node is created.

			Args:
				config: node for which the product is being created
				component_type: if not specified, the type is extracted from the config node with the key "_type"
				modifiers: if not specified, the modifiers are extracted from the config node with the key "_mod"
				project: the owning project, if not specified, the same project as the config is used
				component_entry: if the component entry has already been found, it can be passed here
				silent: if True, suppresses the reporter from printing messages
				**kwargs: additional arguments (unused)
			'''
			if component_type is unspecified_argument:
				component_type = config.pull(self._config_component_key, None, silent=True) \
					if isinstance(config, config.SparseNode) else None
			if component_type is not None and modifiers is None:
				modifiers = config.pull(self._config_modifier_key, None, silent=True)
				if modifiers is None:
					modifiers = []
				elif isinstance(modifiers, dict):
					if len(modifiers):
						key = next(iter(modifiers))
						modifiers = [mod for _, mod in sorted(modifiers.items(), key=lambda x: int(x[0]))] \
							if key.isdigit() \
							else [mod for mod, _ in sorted(modifiers.items(), key=lambda x: (int(x[1]), x[0]))]
				elif isinstance(modifiers, str):
					modifiers = [modifiers]
				elif isinstance(modifiers, (list, tuple)):
					modifiers = list(modifiers)
				else:
					raise ValueError(f'Invalid modifier: {modifiers!r}')
			if project is None:
				project = config.project
			super().__init__(config, **kwargs)
			if project is None:
				prt.warning('No project specified for creator')
			self.silent = silent
			self.project = project
			self.component_type = component_type
			self.modifiers = modifiers
			self.component_entry = component_entry


		def validate(self, config: 'AbstractConfig') -> AbstractCreator:
			'''
			Validates the creator. If the creator is invalid, a new one is created and returned
			based on what is specified in the config or component_entry.

			If a different creator is specified, the current one is replaced with :meth:`DefaultCreator.replace`.
			Otherwise, the creator is returned unchanged.

			Args:
				config: node for which the product is being created

			Returns:
				Validated creator

			'''
			if self.component_type is None:
				return
			if self.component_entry is unspecified_argument:
				self.component_entry = self.project.find_artifact('component', self.component_type)
			creator = config.pull(self._config_creator_key, self.component_entry.creator, silent=True)
			if creator is not None:
				entry = self.project.find_artifact('creator', creator)
				if type(self) != entry.cls:
					return entry.cls.replace(self, config).validate(config)


		@staticmethod
		def _modify_component(component: NamedTuple, modifiers: Optional[Tuple] = ()) -> Type:
			'''
			Modifies the component by applying the given modifiers.

			By default, this will create a subclass of all the modifiers and the original component.

			Args:
				component: entry of the component
				modifiers: entries of the modifiers to apply

			Returns:
				Modified component type

			'''
			cls = component.cls
			if isinstance(cls, AbstractCustomArtifact):
				cls = cls.top
				if len(modifiers) > 0:
					raise ValueError(f'Cannot apply modifiers to custom artifacts: {component.name!r}')
				return cls
			mods = [mod.cls for mod in modifiers]
			if not isinstance(cls, type):
				if len(mods):
					raise ValueError(f'Cannot apply modifiers to non-class components: {component.name!r}')
				assert callable(cls), f'Invalid component: {component.name!r} ({cls!r})'
				return cls
			if issubclass(cls, Modifiable):
				return cls.inject_mods(*reversed(mods))
			# default subclass
			if len(mods):
				bases = (*mods, cls)
				cls = type('_'.join(base.__name__ for base in bases), bases, {})
			return cls


		def _create_component(self, config: 'ConfigNode', args: Tuple, kwargs: Dict[str, Any],
		                      silent: bool = None) -> Any:
			'''
			Creates the component. First finds the component and modifier entries,
			and propagates the silent flag when the component is initialized.

			Note that if the component is a subclass of :class:`AbstractCertifiable`, then after the component
			is fully instantiated and initialized, ``__certify__`` is called with the config
			to certify the component.

			Args:
				config: node for which the product is being created
				args: any manual positional arguments to pass to the component constructor
				kwargs: any manual keyword arguments to pass to the component constructor
				silent: if True, suppresses the reporter from printing messages

			Returns:

			'''
			config.reporter.create_component(config, component_type=self.component_type, modifiers=self.modifiers,
			                                 creator_type=self._creator_name, silent=silent)
			cls = self._modify_component(self.component_entry,
			                             [self.project.find_artifact('modifier', mod) for mod in self.modifiers])

			if isinstance(cls, type) and issubclass(cls, AbstractConfigurable):
				obj = cls.init_from_config(config, args, kwargs, silent=silent)
			else:
				settings = config.settings
				old_silent = settings.get('silent', None)
				settings['silent'] = silent
				obj = cls(config, *args, **kwargs)
				if old_silent is not None:
					settings['silent'] = old_silent

			if isinstance(obj, AbstractCertifiable):
				obj = obj.__certify__(config)

			config._trace = None
			return obj


		def _create_container(self, config: 'ConfigNode', silent: Optional[bool] = None) -> Any:
			'''
			Creates the container, such as a :class:`dict` or :class:`list` based on the config tree structure.

			Depending on what type of node is used, the container will either be a :class:`dict` or :class:`list`.

			All the children of the config node are created using :meth:`ConfigNode.pull`. For this purpose,
			if the current creation is the result of a search, the children are pulled using
			a :meth:`AbstractSearch.sub_search`.

			Args:
				config: node for which the product is being created
				silent: if True, suppresses the reporter from printing messages

			Returns:
				Container with the processed children

			Raises:
				:exc:`NotImplementedError`: if the config node is not recognized


			'''
			config.reporter.create_container(config, silent=silent)
			trace = config.trace

			context = nullcontext() if trace is None else trace.sub_search()

			with context:
				if isinstance(config, config.SparseNode):
					product = {}
					for key, child in config.named_children():
						product[key] = config.pull(key, silent=silent)

				elif isinstance(config, config.DenseNode):
					product = []
					for key, child in config.named_children():
						product.append(config.pull(key, silent=silent))
				else:
					raise NotImplementedError(f'Unknown container type: {type(config)}')

			config._trace = None
			return product


		def _create_primitive(self, config: 'ConfigNode', silent: Optional[bool] = None) -> Any:
			'''
			Creates the primitive, such as a :class:`str` or :class:`int` based on the payload of the config node.

			Args:
				config: node for which the product is being created
				silent: if True, suppresses the reporter from printing messages

			Returns:
				Primitive value contained in the config node

			'''
			payload = config.payload
			config.reporter.create_primitive(config, value=payload, silent=silent)
			config._trace = None
			return payload


		def _setup_context(self, config: 'ConfigNode') -> None:
			'''
			This prepares the creator for creating the node to avoid getting stuck in
			an infinite loop of creating products if there are some cycles in the config tree.

			Args:
				config: node for which the product is being created

			Returns:
				None

			'''
			table = ConfigNode.DefaultCreator._creation_context
			if table is None:
				ConfigNode.DefaultCreator._creation_context = {config: True}
			else:
				if config in table:
					raise config.CycleError(config)
				table[config] = False


		def _end_context(self, config: 'ConfigNode', product: Any) -> None:
			'''
			This cleans up the creator after creating the product.

			Args:
				config: node for which the product is being created
				product: that was created

			Returns:
				None

			'''
			reset = ConfigNode.DefaultCreator._creation_context.get(config, None)
			if reset:
				ConfigNode.DefaultCreator._creation_context = None
			elif reset is not None:
				del ConfigNode.DefaultCreator._creation_context[config]


		def create_product(self, config: 'ConfigNode', args: Optional[Tuple] = None,
		           kwargs: Optional[Dict[str,Any]] = None, *, silent: Optional[bool] = None) -> Any:
			'''
			Top level method for creating the product from the config node which is called by the config node.

			Args:
				config: node for which the product is being created
				args: manual positional arguments to be passed to the component constructor
				kwargs: manual keyword arguments to be passed to the component constructor
				silent: if True, suppresses the reporter from printing messages

			Returns:
				Product created from the config node

			'''
			if args is None:
				args = ()
			if kwargs is None:
				kwargs = {}
			if silent is None:
				silent = self.silent
			
			transfer = self.validate(config)
			if transfer is not None:
				return transfer.create_product(config, args=args, kwargs=kwargs)

			self._setup_context(config)

			if self.component_type is None:
				if config.has_payload:
					value = self._create_primitive(config, silent=silent)
				else:
					value = self._create_container(config, silent=silent)
			else:
				value = self._create_component(config, args=args, kwargs=kwargs, silent=silent)

			self._end_context(config, value)
			return value
			

	def search(self, *queries: str, default: Optional[Any] = AbstractConfig._empty_default, **kwargs) -> Search:
		'''
		Creates a search object that can be used to resolve the given queries to find the corresponding node.

		Note that this method is usually not called directly, but rather through the top level methods such as
		:meth:`ConfigNode.pull` or :meth:`ConfigNode.peek`.

		Args:
			*queries: list of queries to be resolved
			default: if all the queries fail, this value will be returned
			**kwargs: additional keyword arguments to be passed to the search object constructor

		Returns:
			Search object that can be used to traverse the config tree

		'''
		return self.Search(origin=self, queries=queries, default=default, **kwargs)


	def peeks(self, *queries: str, default: Optional[Any] = AbstractConfig._empty_default,
	          silent: Optional[bool] = None) -> 'ConfigNode':
		'''
		Searches in the config based on the given queries and returns the resulting node.

		If multiple queries are given, the first one that resolves to a node will be returned.
		If all the queries fail, the default value will be returned (if given),
		otherwise a :exc:`Search.SearchFailed` will be raised.

		Args:
			*queries: list of queries to be resolved
			default: value to be returned if all the queries fail
			silent: if True, suppresses the reporter from printing messages

		Returns:
			Config node that corresponds to the first query that resolves to a node

		Raises:
			Search.SearchFailed: if all the queries fail and no default value is given

		'''
		return self.search(*queries, default=default).find_node(silent=silent)


	def pulls(self, *queries: str, default: Optional[Any] = AbstractConfig._empty_default,
	          silent: Optional[bool] = None, **kwargs) -> Any:
		'''
		Searches in the config based on the given queries and returns the product of the resulting node.
		The product is the value that node contains, which can be a primitive, a container, or a component.
		If the product is a container or component, it will be created if it has not been created yet.

		If multiple queries are given, the first one that resolves to a node is used
		If all the queries fail, the default value will be returned (if given),
		otherwise a :exc:`Search.SearchFailed` will be raised.

		Args:
			*queries: list of queries to be resolved
			default: value to be returned if all the queries fail
			silent: if True, suppresses the reporter from printing messages

		Returns:
			Product of the config node corresponding to the first query that doesn't fail

		Raises:
			Search.SearchFailed: if all the queries fail and no default value is given

		'''
		return self.search(*queries, default=default, **kwargs).find_product(silent=silent)


	def push(self, addr: str, value: Any, overwrite: bool = True, silent: Optional[bool] = None) -> bool:
		'''
		Inserts a new node into the config tree.

		Args:
			addr: of the new node (i.e. the key or index)
			value: of the new node
			overwrite: if True, overwrites the existing node if there is one
			silent: if True, suppresses the reporter from printing messages

		Returns:
			True if the node was inserted, False if it was not inserted

		Raises:
			ReadonlyError: if the config is readonly

		'''
		if self.settings.get('readonly', False):
			raise self.ReadOnlyError('Cannot push to read-only node')

		if value == self._delete_value:
			self.remove(addr)
			return True

		if not self.has(addr) or overwrite:
			self.set(addr, value)
			return True
		return False


	def __len__(self):
		'''Number of children of the node.'''
		return len(list(self._child_keys()))


	def _child_keys(self) -> Iterator[str]:
		'''Returns the keys of the children of the node.'''
		for key, child in self.named_children():
			if child is not self.empty_value and not child.has_payload or child.payload not in {'__x__', '_x_'}:
				yield key


	def peek_children(self, *, silent: Optional[bool] = None) -> Iterator['ConfigNode']:
		'''
		Returns an iterator over the child nodes of ``self``.

		The iterator skips invalid children (such as empty values or ``__x__``).

		Args:
			silent: if True, suppresses the reporter from printing messages

		Returns:
			Iterator over the children of ``self``

		'''
		for key, child in self.peek_named_children(silent=silent):
			yield child


	def peek_named_children(self, *, silent: Optional[bool] = None) -> Iterator[Tuple[str, 'ConfigNode']]:
		'''
		Returns an iterator over the child nodes of ``self``, including their keys.

		The iterator skips invalid children (such as empty values or ``__x__``).

		Args:
			silent: if True, suppresses the reporter from printing messages

		Returns:
			Iterator producing tuples in the form of (key, child_node)

		'''
		self.reporter.report_iterator(self, product=False, silent=silent)
		for key in self._child_keys():
			child = self.search(key).find_node(silent=silent)
			yield key, child


	def pull_children(self, *, force_create: Optional[bool] = False, silent: Optional[bool] = None) -> Iterator[Any]:
		'''
		Returns an iterator over the products of the child nodes of ``self``.

		The iterator skips invalid children (such as empty values or ``__x__``).

		Args:
			force_create: if True, will always create new products instead of reuse existing ones
			silent: if True, suppresses the reporter from printing messages

		Returns:
			Iterator over the products of the children of the node

		'''
		for key, product in self.pull_named_children(force_create=force_create, silent=silent):
			yield product


	def pull_named_children(self, *, force_create: Optional[bool] = False, silent: Optional[bool] = None) \
			-> Iterator[Tuple[str, Any]]:
		'''
		Returns an iterator over the products of the child nodes of ``self``, including their keys.

		The iterator skips invalid children (such as empty values or ``__x__``).

		Args:
			force_create: if True, will always create new products instead of reuse existing ones
			silent: if True, suppresses the reporter from printing messages

		Returns:
			Iterator over the products of the children of the node

		'''
		for key, child in self.peek_named_children(silent=silent):
			product = child.create() if force_create else child.process()
			yield key, product
		

	def peek_process(self, query, default: Optional[Any] = AbstractConfig._empty_default,
	                 *args: Any, **kwargs: Any) -> Any:
		'''
		Convenience method which composes :meth:`peek` and :meth:`process`, to only process a node if it exists,
		and otherwise return the given default value.

		Args:
			query: of the node to be processed
			default: value to be returned if the node doesn't exist
			*args: positional arguments to be passed to :meth:`process` (e.g. to the constructor of the component)
			**kwargs: keyword arguments to be passed to :meth:`process` (e.g. to the constructor of the component)

		Returns:
			Product of the node corresponding to the query, or the default value if the node doesn't exist

		Raises:
			Search.SearchFailed: if the query fails and no default value is given

		'''
		try:
			node = self.peek(query)
		except self.Search.SearchFailed:
			if default is self._empty_default:
				raise
			return default
		else:
			out = node.process(*args, **kwargs)
			return out


	def peeks_process(self, *queries, default: Optional[Any] = AbstractConfig._empty_default,
	                  **kwargs: Any) -> Any:
		'''
		Convenience method which composes :meth:`peeks` and :meth:`process`, to only process a node if it exists,
		and otherwise return the given default value.

		Note that since this method allows for multiple queries, no positional arguments can be passed to the
		:meth:`process` method, and neither can the keyword argument ``default``.

		Args:
			*queries: of the nodes to be processed
			default: value to be returned if the node doesn't exist
			**kwargs: keyword arguments to be passed to :meth:`process` (e.g. to the constructor of the component)

		Returns:
			Product of the node corresponding to the query, or the default value if the node doesn't exist

		Raises:
			Search.SearchFailed: if the query fails and no default value is given

		'''
		try:
			node = self.peeks(*queries)
		except self.Search.SearchFailed:
			if default is self._empty_default:
				raise
			return default
		else:
			out = node.process(**kwargs)
			return out


	def peek_create(self, query, default: Optional[Any] = AbstractConfig._empty_default,
	                *args: Any, **kwargs: Any) -> Any:
		'''
		Convenience method which composes :meth:`peek` and :meth:`create`,
		to only create the product if the node exists, and otherwise return the given default value.

		Args:
			query: of the node for which the product should be created
			default: value to be returned if the node doesn't exist
			*args: positional arguments to be passed to :meth:`create` (e.g. to the constructor of the component)
			**kwargs: keyword arguments to be passed to :meth:`create` (e.g. to the constructor of the component)

		Returns:
			Product of the node corresponding to the query, or the default value if the node doesn't exist

		Raises:
			Search.SearchFailed: if the query fails and no default value is given

		'''
		try:
			node = self.peek(query)
		except self.Search.SearchFailed:
			if default is self._empty_default:
				raise
			return default
		else:
			out = node.create(*args, **kwargs)
			return out


	def peeks_create(self, *queries, default: Optional[Any] = AbstractConfig._empty_default,
	                  **kwargs: Any) -> Any:
		'''
		Convenience method which composes :meth:`peeks` and :meth:`create`, to only create the product if the node
		exists, and otherwise return the given default value.

		Note that since this method allows for multiple queries, no positional arguments can be passed to the
		:meth:`create` method, and neither can the keyword argument ``default``.

		Args:
			*queries: of the nodes to be processed
			default: value to be returned if the node doesn't exist
			**kwargs: keyword arguments to be passed to :meth:`create` (e.g. to the constructor of the component)

		Returns:
			Product of the node corresponding to the query, or the default value if the node doesn't exist

		Raises:
			Search.SearchFailed: if the query fails and no default value is given

		'''
		try:
			node = self.peeks(*queries)
		except self.Search.SearchFailed:
			if default is self._empty_default:
				raise
			return default
		else:
			out = node.process(**kwargs)
			return out


	def __init__(self, *args, reporter: Optional[Reporter] = None, settings: Optional[Settings] = None,
	             project: Optional[AbstractProject] = None, manager: Optional[AbstractConfigManager] = None,
	             **kwargs):
		'''
		Initializes a new config node. Most of the arguments passed to this constructor should usually be ``None``
		so that this node refers to the values of the root config node.

		Args:
			*args: unused positional arguments passed to the constructor of the super class
			reporter: used for reporting when data is accessed, defers to the root node
			settings: used to determine the behavior of searching and producing products, defers to the root node
			project: associated with this config, defers to the root node
			manager: associated with this config, defers to the root node
			**kwargs: unused keyword arguments passed to the constructor of the super class
		'''
		super().__init__(*args, **kwargs)
		self._project = project
		self._trace = None
		self._product = None
		self._cro = None
		self._bases = None
		self._manager = manager
		self._reporter = reporter
		self._settings = settings
		if self.reporter is None:
			self.reporter = self.Reporter()
		if self.settings is None:
			self.settings = self.Settings()
	
	
	def __deepcopy__(self, memodict={}):
		'''Deep copy of the node (and all subnodes). Does not include the parent node.'''
		new = self.from_raw(self.to_python())
		new._settings = deepcopy(self._settings)
		new._manager = self._manager
		new._project = self._project
		new._trace = self._trace
		new._product = self._product
		new._cro = self._cro
		new._bases = self._bases
		return new
	
	
	def __eq__(self, other):
		'''Compares this config node to another object.'''
		return type(self) == type(other) \
		       and id(self.root) == id(other.root) \
		       and self.my_address() == other.my_address()


	def __hash__(self):
		'''Returns a hash value for this config node.'''
		return hash(self.my_address())


	def export(self, name: Union[str, Path], *, root: Optional[Union[str, Path]] = None,
			   fmt: Optional[str] = None) -> Optional[Path]:
		'''
		Exports the given config to the given path (in yaml format).

		Args:
			config: object to export
			name: of file name or path to export to (without extension)
			root: directory to export to (if not provided, the current working directory is used)
			fmt: format to export to (if not provided, the extension of the file name is used, and defaults to yaml)

		Returns:
			The path to which the config was exported

		'''
		return self.manager.export(self, name, root=root, fmt=fmt)


	@property
	def project(self):
		'''Returns the project associated with this config tree.'''
		if self._project is None:
			parent = self.parent
			if parent is not None:
				return parent.project
		return self._project
	@project.setter
	def project(self, project: AbstractProject):
		parent = self.parent
		if parent is None:
			self._project = project
		else:
			parent.project = project


	@property
	def cro(self) -> Tuple[str, ...]:
		'''
		Returns the list of all config files that were composed to produce this config tree.
		Analogous to the method resolution order (``mro``) for classes.
		'''
		if self._cro is None:
			if self.parent is None:
				return ()
			return self.parent.cro
		return self._cro


	@property
	def bases(self) -> Tuple[str, ...]:
		'''
		Returns the list of config files that were explicitly mentioned to produce this config tree.
		Analogous to ``__bases__`` for classes.
		'''
		if self._bases is None:
			if self.parent is None:
				return ()
			return self.parent.bases
		return self._bases


	@property
	def manager(self):
		'''Returns the manager associated with this config tree.'''
		if self._manager is None:
			parent = self.parent
			if parent is not None:
				return parent.manager
		return self._manager
	@manager.setter
	def manager(self, manager: AbstractConfigManager):
		parent = self.parent
		if parent is None:
			self._manager = manager
		else:
			parent.manager = manager


	@property
	def trace(self) -> Optional[Search]:
		'''
		Returns the current search trace associated this config node
		(generally only used by creators and reporters).
		'''
		return self._trace


	@property
	def reporter(self) -> Reporter:
		'''Returns the reporter associated with this config tree.'''
		if self._reporter is None:
			parent = self.parent
			if parent is not None:
				return parent.reporter
		return self._reporter
	@reporter.setter
	def reporter(self, reporter: Reporter):
		parent = self.parent
		if parent is None:
			self._reporter = reporter
		else:
			parent.reporter = reporter


	@property
	def settings(self) -> Settings:
		'''Returns the (global) settings associated with this config tree.'''
		if self._settings is None:
			parent = self.parent
			if parent is not None:
				return parent.settings
		return self._settings
	@settings.setter
	def settings(self, settings: Settings):
		parent = self.parent
		if parent is None:
			self._settings = settings
		else:
			parent.settings = settings


	@property
	def silent(self) -> bool:
		'''Returns whether this config node is in silent mode.'''
		return self.settings.get('silent', False)
	@silent.setter
	def silent(self, value: bool):
		'''Sets whether this config node is in silent mode.'''
		self.settings['silent'] = value


	class ReadOnlyError(Exception):
		'''Raised when a read-only config node is attempted to be modified.'''
		pass


	class ConfigContext:
		'''
		Context manager for temporarily modifying the (global) settings of a config tree.
		'''
		def __init__(self, config: 'ConfigNode', settings: Dict[str, bool]):
			self.config = config
			self.old_settings = None
			self.settings = settings

		def __enter__(self):
			settings = self.config.settings
			self.old_settings = settings.copy()
			settings.update(self.settings)

		def __exit__(self, exc_type, exc_val, exc_tb):
			settings = self.config.settings
			settings.clear()
			settings.update(self.old_settings)
	def context(self, **settings: bool) -> ContextManager:
		'''Returns a context manager for temporarily modifying the (global) settings of this config tree.'''
		return self.ConfigContext(self, settings)
		

	def _create(self, component_args: Optional[Tuple] = None, component_kwargs: Optional[Dict[str,Any]] = None,
	            silent: Optional[bool] = None, creator: Optional[str] = unspecified_argument, **kwargs: Any) -> Any:
		'''
		Creates the product of this config node (which may be a primitive, a container such as a list or dict,
		or a component).

		Args:
			component_args: positional arguments to pass to the product constructor
			component_kwargs: keyword arguments to pass to the product constructor
			silent: if True, suppresses the reporter from printing any messages
			creator: the creator to use to create the product (if unspecified, the "creator" setting is checked,
			or the default creator is used)
			**kwargs: additional keyword arguments to pass to the creator for it's own initialization

		Returns:
			The newly created product of this config node.

		'''
		if creator is unspecified_argument:
			creator = self.settings.get('creator')
		creator = self.DefaultCreator if creator is None else self.project.find_artifact('creator', creator).cls
		out = creator(self, silent=silent, project=self.project,  **kwargs)\
			.create_product(self, args=component_args, kwargs=component_kwargs, silent=silent)
		return out


	def _process(self, component_args: Optional[Tuple] = None, component_kwargs: Optional[Dict[str, Any]] = None,
	             silent: Optional[bool] = None, **kwargs: Any) -> Any:
		'''
		Returns the product of this config node. If the product of the node has already been created, that is returned,
		otherwise the product is created using :meth:`_create`.

		Args:
			component_args: positional arguments to pass to the product constructor (if the product is created)
			component_kwargs: keyword arguments to pass to the product constructor (if the product is created)
			silent: if True, suppresses the reporter from printing any messages (if the product is created)
			**kwargs: additional keyword arguments to pass to the creator for it's own initialization
			(if the product is created)

		Returns:
			The product of this config node.

		'''
		settings = self.settings
		force_create = settings.get('force_create', False)
		allow_create = settings.get('allow_create', True)
		assert not (force_create and not allow_create), f'Cannot force create without allowing create: {self}'
		if (allow_create and self._product is None) or force_create:
			self._product = self._create(component_args, component_kwargs, silent=silent, **kwargs)
		else:
			self.reporter.reuse_product(self, self._product, silent=silent)
		return self._product


	def create(self, *args: Any, **kwargs: Any) -> Any:
		'''
		Creates a new value based on the contents of self.

		Args:
			*args: Manual arguments to pass to the value constructor.
			**kwargs: Manual keyword arguments to pass to the value constructor.

		Returns:
			The newly created value.

		'''
		return self._create(args, kwargs, silent=self.settings.get('silent', None))


	def create_silent(self, *args: Any, **kwargs: Any) -> Any:
		'''Convenience method for creating a value in silent mode.'''
		return self._create(args, kwargs, silent=True)


	def process(self, *args: Any, **kwargs: Any) -> Any:
		'''
		Processes the config object using the contents of self.

		If a value for this config object has already been created, it is returned instead of creating a new one.

		Args:
			*args: Manual arguments to pass to the value constructor. (ignored if a value has already been created)
			**kwargs: Manual keyword arguments to pass to the value constructor.
			(ignored if a value has already been created)

		Returns:
			The processed value.

		'''
		return self._process(args, kwargs, silent=self.settings.get('silent', None))


	def process_silent(self, *args: Any, **kwargs: Any) -> Any:
		'''Convenience method for processing a value in silent mode.'''
		return self._process(args, kwargs, silent=True)


	@property
	def product_exists(self) -> bool:
		'''Returns True if the product of this config node has already been created.'''
		return self._product is not None


	def clear_product(self, recursive: bool = True) -> None:
		'''
		Removes the product of this config node (if it exists), and optionally removes the products of all child nodes.

		Args:
			recursive: if True, the products of all child nodes are also removed

		'''
		self._product = None
		if recursive:
			for _, child in self.named_children():
				child.clear_product(recursive=recursive)

	def to_yaml(self, stream=None, default_flow_style=None, sort_keys=True, **kwargs: Any) -> None:
		'''
		Dumps the contents of this config node in YAML format to the specified stream.

		Args:
			stream: destination stream
			default_flow_style: YAML formatting option (see PyYAML documentation)
			sort_keys: YAML formatting option (if True, keys are sorted alphabetically)
			**kwargs: additional keyword arguments to pass to PyYAML's dump function

		Returns:
			None

		'''
		return yaml.dump(self.to_python(), stream, default_flow_style=default_flow_style, sort_keys=sort_keys,
		                 **kwargs)


	def __str__(self):
		'''Returns a string representation of the contents of this config node.'''
		return f'<{self.payload!r}>' if self.is_leaf else self.to_yaml()


	def __repr__(self):
		'''Returns a string representation of this config node.'''
		return f'<{self.__class__.__name__} {len(self)} children>'


	def update(self, update: 'ConfigNode', *, clear_product: bool = True) -> 'ConfigNode':
		'''
		Updates the contents of this config node with the contents of another config node.

		Args:
			update: the config node to update from
			clear_product: if True, all references to the products of ``self`` and ``update`` are removed

		Returns:
			The updated config node, ``self``

		'''
		if clear_product:
			self.clear_product()
			update.clear_product()
		if update.has_payload:
			self.payload = update.payload
		elif self.has_payload:
			self.payload = unspecified_argument
		for key, child in update.named_children():
			child.parent = self
			if key in self:
				self[key].update(child)
			else:
				self[key] = child
		return self


	_delete_value = '_x_'
	def validate(self):
		'''
		Recursively validates the contents of this config node.

		Including removing any children with the value ``_x_`` (marked for removal).
		'''
		bad = []
		for key, child in self.named_children():
			if child.has_payload and child.payload == self._delete_value:
				bad.append(key)
			else:
				child.validate()
		for key in bad:
			self.remove(key)

	def silence(self, silent: bool = True) -> ContextManager:
		'''Convenience method for temporarily setting the silent flag of this config node.'''
		return self.context(silent=silent)

	def print(self, *terms: Any, force: bool = False, sep: str = ' ', end: str = '\n', **kwargs):
		'''Convenience method for printing iff (`force` or not self.silent)'''
		if force or not self.silent:
			print(*terms, sep=sep, end=end, **kwargs)



class ConfigSparseNode(AutoTreeSparseNode, ConfigNode):
	'''A config node that treats its children as being in a dict.'''
	_python_structure = dict

	def _get(self, addr: str):
		try:
			return super()._get(addr)
		except self._MissingKey:
			try:
				return super()._get(addr.replace('-', '_'))
			except self._MissingKey:
				pass
			raise

	def _set(self, addr: str, node):
		return super()._set(addr, node)

	def _remove(self, addr: str):
		if addr in self._children:
			del self._children[addr]

	def _has(self, addr: str):
		return addr in self._children or addr.replace('-', '_') in self._children





class ConfigDenseNode(AutoTreeDenseNode, ConfigNode):
	'''A config node that treats its children as being in a list.'''
	_python_structure = list



ConfigNode.DefaultNode = ConfigSparseNode
ConfigNode.SparseNode = ConfigSparseNode
ConfigNode.DenseNode = ConfigDenseNode










from ._info import __version__, __author__, __info__, __logger__

from .organization import get_profile, ProfileBase, ProjectBase, GeneralProject, Profile
from .registration import script, component, modifier, creator, autocomponent, autoscript
from .top import get_current_project, get_project, switch_project, iterate_projects, \
	entry, main, run_script, run, quick_run, initialize, create_config, parse_argv

from .config import ConfigNode as Configuration
from . import exporting as _exporting
from .behaviors import Behavior

from .configurable import Configurable, Certifiable, config_aliases, silent_config_args




from typing import List, Dict, Optional, Union, Any, Sequence, Iterator, NamedTuple
import os
from pathlib import Path
import io
import yaml
from collections import OrderedDict
from omnibelt import Path_Registry, JSONABLE, unspecified_argument, export, load_export, linearize, CycleDetectedError

from ..abstract import AbstractConfig, AbstractProject, AbstractConfigManager
from .nodes import ConfigNode



class ConfigManager(AbstractConfigManager):
	_config_path_delimiter = '/'
	_config_exts = ('yaml', 'yml', 'json', 'tml', 'toml')
	_config_nones = {'None', 'none', '_none', '_None', 'null', 'nil', }
	_config_parent_key = '_base'

	ConfigNode = ConfigNode


	class Config_Registry(Path_Registry, components=['project']):
		'''Registry for config files.'''
		pass


	def __init__(self, project: AbstractProject):
		self.registry = self.Config_Registry()
		self.project = project


	def export(self, config: ConfigNode, name: Union[str, Path], *, root: Optional[Path] = None,
			   fmt: Optional[str] = None) -> Path:
		'''
		Exports the given config to the given path (in yaml format).

		Args:
			config: object to export
			name: of file to export to
			root: directory to export to (if not provided, the current working directory is used)
			fmt: format to export to (if not provided, the extension of the file name is used, and defaults to yaml)

		Returns:
			The path to which the config was exported

		'''
		return export(config, path=name, root=root, fmt=fmt)


	def iterate_configs(self) -> Iterator[NamedTuple]:
		'''
		Iterates over all registered config file entries.

		Returns:
			An iterator over all registered config file entries.

		'''
		yield from self.registry.values()


	def register_config(self, name: Union[str, Path], path: Union[str, Path] = None, **kwargs) -> NamedTuple:
		'''
		Registers a config file with the given name.

		Note:
			It is generally not recommended to register configs manually, but rather to use the ``register_config_dir``
			method to register all configs in a directory at once.

		Args:
			name: to register the config under
			path: of the config file (if not provided, the provided name is assumed to be a path)
			**kwargs: Other arguments to pass to the ``Config_Registry.register`` method

		Returns:
			The entry of the config file that was registered

		'''
		if path is None:
			path = name
			name = None
		if name is None:
			name = Path(path).stem
		return self.registry.new(name, path, **kwargs)


	def register_config_dir(self, root: Union[str, Path], *, recursive: Optional[bool] = True,
	                        prefix: Optional[str] = None, delimiter: Optional[str] = None) -> List[NamedTuple]:
		'''
		Registers all yaml files found in the given directory (possibly recursively)

		When recusively checking all directories inside, the internal folder hierarchy is preserved
		in the name of the config registered, so for example if the given ``path`` points to a
		directory that contains a directory ``a`` and two files ``f1.yaml`` and ``f2.yaml``:

		Contents of ``path`` and corresponding registered names:

			- ``f1.yaml`` => ``f1``
			- ``f2.yaml`` => ``f2``
			- ``a/f3.yaml`` => ``a/f3``
			- ``a/b/f4.yaml`` => ``a/b/f3``

		If a ``prefix`` is provided, it is appended to the beginning of the registered names

		Args:
			root: root directory to search through
			recursive: search recursively through subdirectories for more config yaml files
			prefix: prefix for names of configs found herein
			delimiter: string to merge directories when recursively searching (default ``/``)

		Returns:
			A list of all config entries that were registered.

		'''
		if delimiter is None:
			delimiter = self._config_path_delimiter
		if prefix is None:
			prefix = ''
		root = Path(root)
		entries = []
		for ext in self._config_exts:
			for path in sorted(root.glob(f'**/*.{ext}' if recursive else f'*.{ext}')):
				terms = path.relative_to(root).parts[:-1]
				name = path.stem
				ident = prefix + delimiter.join(terms + (name,))
				entries.append(self.register_config(ident, path, project=self.project))
		return entries

	def _parse_raw_arg(self, arg: str) -> JSONABLE:
		val = yaml.safe_load(io.StringIO(arg))
		if isinstance(val, str) and val in self._config_nones:
			return None
		return val


	class UnknownBehaviorError(ValueError):
		'''While parsing command-line arguments, a meta config was referenced that was not registered'''
		def __init__(self, term):
			super().__init__(f'{term!r} was not recognized (and consequently removed) by any registered Behaviors.')


	def parse_argv(self, argv: Sequence[str], script_name: Optional[str] = unspecified_argument) -> AbstractConfig:
		'''
		Parses command-line arguments and returns a config object that contains the parsed arguments.

		Args:
			argv: raw command-line arguments
			script_name: if not provided, the script name is inferred from ``argv``

		Returns:
			A config object containing the parsed arguments

		Raises:
			UnknownBehaviorError: if an unknown behavior config is encountered
			ValueError: if an argument is invalid

		'''

		meta = {}
		argv = list(argv)

		for behavior in self.project.behaviors():
			out = behavior.parse_argv(meta, argv, script_name=script_name)
			if out is not None:
				argv = out

		configs = []
		data = {}

		remaining = list(reversed(argv))
		if len(remaining):
			# parse script (if provided)
			term = remaining.pop()
			if term.startswith('-') and not term.startswith('--'):
				raise self.UnknownBehaviorError(term)
			elif term.startswith('--'):
				remaining.append(term)
			elif script_name is not unspecified_argument:
				remaining.append(term)
			elif term != '_' and script_name is unspecified_argument:
				script_name = term
			if script_name not in {None, unspecified_argument}:
				meta['script_name'] = script_name

			# parse remaining (keyword) arguments
			while len(remaining):
				term = remaining.pop()

				if term.startswith('--'):
					remaining.append(term)
					break
				else:
					configs.append(term)

			waiting_arg_key = None
			while len(remaining):
				term = remaining.pop()

				if term.startswith('--'):
					if waiting_arg_key is not None:
						data[waiting_arg_key] = True

					key, *other = term[2:].split('=', 1)
					if len(other) and len(other[0]):
						if len(other) > 1:
							raise ValueError(f'Invalid argument {term} (avoid using "=" in argument names)')
						data[key] = self._parse_raw_arg(other[0])
					else:
						waiting_arg_key = key

				elif waiting_arg_key is not None:
					data[waiting_arg_key] = self._parse_raw_arg(term)
					waiting_arg_key = None

				else:
					raise ValueError(f'Unexpected argument: {term}')

			if waiting_arg_key is not None:
				data[waiting_arg_key] = True

		# configurize parsed data (including meta)
		config = self.create_config(configs, data)
		meta = self.create_config(None, meta)
		meta_base = config.push_peek('_meta', {}, silent=True, overwrite=False)
		meta_base.update(meta)
		return config


	def find_local_config_entry(self, name: str, default: Optional[Any] = unspecified_argument) -> NamedTuple:
		'''
		Finds the entry for a config by name.

		Args:
			name: used to register the config
			default: Default value to return if the artifact is not found.

		Returns:
			The entry for the config

		Raises:
			ConfigNotFoundError: if the config is not registered

		'''
		try:
			return self.registry.find(name)
		except self.registry.NotFoundError:
			pass
		if default is not unspecified_argument:
			return default
		raise self.ConfigNotRegistered(name)


	def find_project_config_entry(self, name: str, default: Optional[Any] = unspecified_argument) -> NamedTuple:
		'''
		Finds the entry for a config by name. Including checking the nonlocal projects.

		Args:
			name: of the config file used to register the config
			default: Default value to return if the artifact is not found.

		Returns:
			The entry for the config

		Raises:
			ConfigNotFoundError: if the config is not registered

		'''
		return self.project.find_artifact('config', name, default=default)


	def find_config_path(self, name: str, default: Optional[Any] = unspecified_argument) -> Path:
		'''
		Finds the path to a config file by name. Including checking the nonlocal projects.

		Args:
			name: of the config file used to register the config
			default: Default value to return if the artifact is not found.

		Returns:
			The path to the config file

		Raises:
			ConfigNotFoundError: if the config is not registered

		'''
		if isinstance(name, Path) and name.is_file():
			return name
		elif os.path.isfile(name):
			return Path(name)
		try:
			return self.find_project_config_entry(name).path
		except self.ConfigNotRegistered:
			if default is unspecified_argument:
				raise
			return default


	@classmethod
	def _find_config_parents(cls, path: Optional[Path], raw: Dict[str, Any]) -> List[str]:
		'''
		Finds the base configs of a config file.

		By default, the ``path`` is not used, and instead ``raw`` is checked for a ``_base`` key.

		Args:
			path: of the config file
			raw: loaded data of the config file

		Returns:
			Names of the base configs of the config file (in order)

		'''
		src = []
		if raw is None:
			return src
		if cls._config_parent_key is not None:
			base = raw.get(cls._config_parent_key, [])
			if isinstance(base, str):
				base = [base]
			src.extend(base)
		assert isinstance(src, list), f'Invalid parents: {src}'
		return src


	def _merge_raw_configs(self, raws: List[JSONABLE]) -> AbstractConfig:
		'''
		Merges a list of raw configs into a single config object.

		Args:
			raws: list of raw configs (in order)

		Returns:
			The merged config

		'''
		singles = [self.configurize(raw) for raw in raws]

		if not len(singles):
			return self.ConfigNode.from_raw({})

		merged = singles.pop()
		while len(singles):
			update = singles.pop()
			merged.update(update)

		return merged


	def load_raw_config(self, path: Union[str, Path]) -> JSONABLE:
		'''
		Loads raw data of a config file (formats: JSON, YAML, TOML).

		Args:
			path: to the config file

		Returns:
			The raw data of the config file

		Raises:
			ValueError: if the config file is not a valid format

		'''
		if path.suffix.endswith('.yml') or path.suffix.endswith('.yaml'):
			# return load_yaml(path, ordered=True)
			return load_export(path=path, fmt='yaml', ordered=True)
		elif path.suffix == '.json':
			# return load_json(path)
			return load_export(path=path, fmt='json')
		elif path.suffix in ('.toml', '.tml'):
			# return toml.load(path, _dict=OrderedDict)
			return load_export(path=path, fmt='toml', _dict=OrderedDict)
		raise ValueError(f'Unknown config file type: {path}')


	class ConfigCycleError(ValueError):
		'''
		Indicates that a cycle in the config bases was detected.
		'''
		def __init__(self, bad: List[str]):
			super().__init__(f'Config cycle detected within: {", ".join(bad)}')


	def create_config(self, configs: Optional[Sequence[Union[str, Path]]] = None, data: Optional[JSONABLE] = None, *,
	                  project: Optional[AbstractProject] = unspecified_argument) -> AbstractConfig:
		'''
		Creates a config object from a list of configs and raw data.

		Args:
			configs: names of registered configs or paths to config files to load
			data: raw data to merge into the config (e.g. from command line arguments)
			project: to associate the resulting config with

		Returns:
			The config object

		Raises:
			ConfigNotFoundError: if a requested config is not registered

		'''
		if configs is None:
			configs = []
		if data is None:
			data = {}
		if project is unspecified_argument:
			project = self.project
		configs = list(configs) + list(self._find_config_parents(None, data))
		# assert len(self._find_config_parents(None, data)) == 0, 'Passed in args cannot have a parents key'

		ancestry_names = {}
		parent_table = {None: configs}
		raws = {None: data}
		used_paths = {}

		todo = list(configs)
		while len(todo):
			name = todo.pop()
			path = self.find_config_path(name)
			if path not in raws:
				if not path.exists():
					raise FileNotFoundError(path)
				used_paths[name] = path
				raws[path] = self.load_raw_config(path)
				if raws[path] is None:
					raws[path] = {}
				ancestry_names[path] = name
				parent_table[path] = self._find_config_parents(path, raws[path])
				todo.extend(parent_table[path])

		if len(used_paths) != len(todo):
			graph = {key: [used_paths[name] for name in srcs] for key, srcs in parent_table.items()}

			try:
				order = linearize(graph)[None]
			except CycleDetectedError as c:
				bad = c.remaining
				del bad[None]
				used_names = {path: name for name, path in used_paths.items()}
				bad = [used_names[path] for path in bad]
				raise self.ConfigCycleError(sorted(bad))

			ancestry = [ancestry_names[p] for p in order[1:]]
			order = [data] + [raws[p] for p in order[1:]]
		else:
			order = [data]
			ancestry = []

		merged = self._merge_raw_configs(order)

		merged._cro = tuple(map(str,ancestry))
		merged._bases = tuple(map(str,configs))

		if project is not None:
			merged.project = project
		merged.settings = merged.pull('_meta.settings', {}, silent=True)

		merged.validate()
		merged.manager = self
		return merged


	def configurize(self, raw: JSONABLE, **kwargs: Any) -> AbstractConfig:
		'''
		Converts raw data into a config object (using the config class of this manager ``ConfigNode``).

		Args:
			raw: python data structure to convert (e.g. from a JSON or YAML file)
			**kwargs: additional arguments to pass to the config class constructor

		Returns:
			The config object

		'''
		return self.ConfigNode.from_raw(raw, **kwargs)


	def merge_configs(self, *configs: AbstractConfig) -> AbstractConfig:
		'''
		Given a list of config objects, merges them into a single config object in the given order.

		Args:
			*configs: objects to merge

		Returns:
			The merged config object

		'''
		merged = self.create_config()
		
		todo = list(configs)
		while len(todo):
			self.update_config(merged, todo.pop())
		return merged


	@staticmethod
	def update_config(base: AbstractConfig, update: AbstractConfig) -> AbstractConfig:
		'''
		Updates a config object with the contents of another config object.

		Args:
			base: config object to update
			update: config object to merge into the base

		Returns:
			The updated config object

		'''
		return base.update(update)
		










from typing import List, Dict, Tuple, Optional, Union, Any, Sequence, Callable, Type
from inspect import Parameter
from omnibelt import extract_function_signature

from .abstract import AbstractCreator, AbstractConfig, AbstractProject, AbstractCustomArtifact
from .top import get_current_project



class _Registration_Decorator:
	'''Base class for all registration decorators'''
	def __init__(self, name: Optional[str] = None, **kwargs: Any):
		'''

		Args:
			name: name of item to be registered (defaults to its __name__)
			**kwargs: additional keyword arguments to pass to :func:`register_script()`
		'''
		self.name = name
		self.kwargs = kwargs
		self.item = None


	def __call__(self, item: Callable) -> Callable:
		'''Decorator call that registers the item. Leaves the decorated item unchanged.'''
		if self.name is None:
			self.name = item.__name__
		self.item = item
		self.register(self.name, item, **self.kwargs)
		return item


	@staticmethod
	def register(name: str, item: Callable[[AbstractConfig], Any], **kwargs) -> None:
		'''Must be implemented by subclasses to register the item'''
		raise NotImplementedError



class _Project_Registration_Decorator(_Registration_Decorator):
	'''Registration decorator which registers the item with the current project'''


	@classmethod
	def register(cls, name: str, item: Callable[[AbstractConfig], Any], project: Optional[AbstractProject] = None,
	             **kwargs) -> None:
		'''Registers the item with the current project using :func:`get_current_project()`.'''
		if project is None:
			project = get_current_project()
		cls.register_project(project, name, item, **kwargs)


	@staticmethod
	def register_project(project: AbstractProject, name: str, item: Callable[[AbstractConfig], Any],
	                     **kwargs) -> None:
		'''Must be implemented by subclasses to register the item with the current project'''
		raise NotImplementedError



class script(_Project_Registration_Decorator):
	'''
	Decorator to register a script.

	Scripts are callable objects (usually functions) with only one input argument (the config object) and can be called
	from the command line using the :code:`fig` command.
	'''

	def __init__(self, name: Optional[str] = None, description: Optional[str] = None, *,
	             hidden: bool = None) -> None:
		'''

		Args:
			name: name of item to be registered (defaults to its __name__)
			description: a short description of what the script does (defaults to first line of its docstring)
			hidden: if True, the script will not be listed in the help menu
		'''
		super().__init__(name=name, description=description, hidden=hidden)


	@staticmethod
	def register_project(project: AbstractProject, name: str, item: Callable[[AbstractConfig], Any],
	                     description: Optional[str] = None, hidden: Optional[bool] = None, **kwargs) -> None:
		if description is None and item.__doc__ is not None:
			description = item.__doc__.split('\n')[0]
		if hidden is None:
			hidden = name.startswith('_')
		project.register_artifact('script', name, item, description=description, hidden=hidden, **kwargs)



class creator(_Project_Registration_Decorator):
	'''
	Decorator to register a creator.

	Creators are generally subclasses of :class:`AbstractCreator` and are used to create objects from the config.

	Usually, the default creator is sufficient, but this decorator can be used to register a custom creator.
	'''

	def __init__(self, name: Optional[str] = None, description: Optional[str] = None):
		'''

		Args:
			name: name of item to be registered (defaults to its __name__)
			description: a short description of what the script does (defaults to first line of its docstring)
		'''
		super().__init__(name=name, description=description)


	@staticmethod
	def register_project(project: AbstractProject, name: str, item: Callable[[AbstractConfig], Any],
	                     description: Optional[str] = None, **kwargs) -> None:
		item._creator_name = name
		project.register_artifact('creator', name, item, description=description, **kwargs)



class component(_Project_Registration_Decorator):
	'''
	Decorator to register a component.

	Components are (usually) classes, and can be automatically be instantiated from the config object
	(using the ``_type`` key).

	There are generally two different ways to use components. Both use a creator (see :class:`AbstractCreator`):
		1. If the component is a subclass of :class:`Configurable`,
			arguments in __init__ can be automatically be filled in with the config object.
		2. Otherwise, the component will be instantiated (by default) with the following signature:
			:code:`config, *args, **kwargs`, where :code:`config` is the config object,
			while :code:`*args` and :code:`**kwargs` are arguments manually passed to the creator.
			This is the signature expected for :func:`init_from_config()` if the component
			is a subclass of :class:`AbstractConfigurable` and :func:`__init__` otherwise.
	'''

	def __init__(self, name: Optional[str] = None, description: Optional[str] = None, creator: Optional[str] = None):
		'''
		Decorator to register a component.

		Args:
			name: name of item to be registered (defaults to its __name__)
			description: a short description of what the script does (defaults to first line of its docstring)
			creator: name of the creator that should be used to create this component (generally not recommended)
		'''
		super().__init__(name=name, creator=creator, description=description)


	@staticmethod
	def register_project(project: AbstractProject, name: str, item: Callable[[AbstractConfig], Any],
	                     description: Optional[str] = None, creator: Optional[Union[str, AbstractCreator]] = None,
	                     **kwargs) -> None:
		if description is None and item.__doc__ is not None:
			description = item.__doc__.split('\n')[0]
		project.register_artifact('component', name, item, description=description, creator=creator, **kwargs)



class modifier(_Project_Registration_Decorator):
	'''
	Decorator to register a modifier.

	Modifiers are "runtime mixins" for components and must be classes. When specifying a component to be modified
	with the ``_mod`` key in the config, a new type is dynamically created for which the bases are all the specified
	modifiers followed by the original component.
	'''

	def __init__(self, name: Optional[str] = None, description: Optional[str] = None):
		'''
		Decorator to register a modifier.

		Args:
			name: name of item to be registered (defaults to its __name__)
			description: a short description of what the script does (defaults to first line of its docstring)
		'''
		super().__init__(name=name, description=description)


	@staticmethod
	def register_project(project: AbstractProject, name: str, item: Callable[[AbstractConfig], Any],
	                     description: Optional[str] = None, **kwargs) -> None:
		if description is None and item.__doc__ is not None:
			description = item.__doc__.split('\n')[0]
		project.register_artifact('modifier', name, item, description=description, **kwargs)



class _AutofillMixin(_Registration_Decorator, AbstractCustomArtifact):
	'''Mixin for decorators that autofill arguments from config'''

	def __init__(self, name: Optional[str] = None,
	             aliases: Optional[Dict[str,Union[str,Sequence[str]]]] = None, **kwargs):
		'''
		Shared constructor for decorators that autofill arguments from config.

		Args:
			name: name of item to be registered (defaults to its __name__)
			aliases: alternative names for arguments (can have multiple aliases per argument)
			**kwargs: additional keyword arguments to pass to :func:`register_script()`
		'''
		if aliases is None:
			aliases = {}
		super().__init__(name=name, **kwargs)
		self.aliases = aliases


	def get_wrapped(self) -> Union[Callable, Type]:
		return self.item


	def autofill(self, config: AbstractConfig, args: Optional[Tuple] = None, kwargs: Optional[Dict[str, Any]] = None) \
			-> Tuple[List[Any], Dict[str, Any]]:
		'''
		Autofill arguments needed for the original item (which was decorated) from config.

		Args:
			config: Config object to autofill from
			args: Manually specified arguments
			kwargs: Manually specified keyword arguments

		Returns:
			Arguments to pass to the original item
		'''
		def default_fn(key, default):
			if default is Parameter.empty:
				default = config._empty_default
			aliases = self.aliases.get(key, ())
			if isinstance(aliases, str):
				aliases = (aliases,)
			return config.pulls(key, *aliases, default=default)
		return extract_function_signature(self.item, args=args, kwargs=kwargs, default_fn=default_fn)


	def top(self, config: AbstractConfig, *args: Any, **kwargs: Any) -> Any:
		'''
		Replacement item to be registered, which first autofills arguments
		from the config and then calls the original item.

		Args:
			config: Config object to autofill from
			*args: Manually specified arguments
			**kwargs: Manually specified keyword arguments

		Returns:
			Result of calling the original item
		'''
		fixed_args, fixed_kwargs = self.autofill(config, args=args, kwargs=kwargs)
		return self.item(*fixed_args, **fixed_kwargs)


	def register(self, name: str, item: Callable[[Any], Any], **kwargs):
		super().register(name, self, **kwargs)



class autoscript(_AutofillMixin, script):
	'''
	Convienence decorator to register scripts where the arguments of the script signature
	are automatically extracted from the config before running the script.

	Note:
		This is generally only recommended for simple, short scripts (since it severely limits the usage of the
		config object by the script).
	'''

	def __init__(self, name: Optional[str] = None, description: Optional[str] = None,
	             aliases: Optional[Dict[str, Union[str, Sequence[str]]]] = None, **kwargs):
		'''
		Decorator to register a script (where arguments are extracted from the config automatically).

		Args:
			name: name of item to be registered (defaults to its __name__)
			description: a short description of what the script does (defaults to first line of its docstring)
			aliases: alternative names for arguments (can have multiple aliases per argument)
		'''
		super().__init__(name, description=description, aliases=aliases, **kwargs)



class autocomponent(_AutofillMixin, component):
	'''
	Convienence decorator to register components where the arguments of the component function
	are automatically extracted from the config

	Note:
		This is generally only recommended for simple components that are functions (rather than classes),
		since class components should simply subclass :class:`Configurable` for effectively the same behavior.
	'''

	def __init__(self, name: Optional[str] = None, description: Optional[str] = None,
	             aliases: Optional[Dict[str, Union[str, Sequence[str]]]] = None,
	             creator: Optional[Union[str, AbstractCreator]] = None):
		'''
		Decorator to register a component (where arguments are extracted from the config automatically).

		Args:
			name: name of item to be registered (defaults to its __name__)
			description: a short description of what the script does (defaults to first line of its docstring)
			aliases: alternative names for arguments (can have multiple aliases per argument)
			creator: name of the creator that should be used to create this component
		'''
		super().__init__(name=name, creator=creator, description=description, aliases=aliases)
	


